// This source file deals with the agents

// NB: on IRIX, Solaris and Linux/glibc realloc() leaves unchanged the block if failure

#include <stdlib.h>
#include <stdio.h>
#include <string.h>     // for memcpy
#include "agents.h"
#include "intern.h"
#include "lowinterface.h"
#include "nana.h"

#define INCDEAD 100     // chunk for deadagents to allocate by
#define INCKILLED 100   // chunk for global killed agents to allocate by
#define INCLOCALCA 100  // chunk to increment local created ag structure by
#define INCLOCALKA 100  // chunk to increment local killed ag structure by
static stack_t deadAgents;  // the holes (dead agents) in agent structure

// agent structure
typedef struct{
  pmpfAgent func;
  int slowness;
  int priority;
  int load, maxload;    // useful only for sources
  int y, x;		// actual position
  int newy, newx;       // planned position
  int actiontype;       // the type of the action to do
  int actionparam;      // param of actiontype (value if TAKE, DROP).  This field can disappear if TAKE, DROP are executed in planMove, not in doMove
  void *memory;         // memory, allocated and freed by the library
  int isKilled;         // this agent will by killed this cycle
  int isAlive;          // set to 0 when this agent dies
#ifndef WITHOUT_NANA
  int plan, move, migr; // check that every agent is planified/moved once per cycle
#endif
}agent_t;
#define INCAG	10	// allocates by INCAG the agent structure
static agent_t *agents;

static int nbAg;        // actual (real) number of agents (nbAg + nbholes = nbmaxAg)
static int nbmaxAg;     // the max value used in agents structure
static int nbAllocAg;   // number of allocated agents
static int nbEmptyAg;   // (dynamic) number of empty-load agents
static int nbRandomAg;  // (dynamic) number of random agents (all agents MOVERANDOM)

#ifndef WITHOUT_NANA
void clearAgents (void)
{
  int ag;
  for (ag=0 ; ag<nbmaxAg ; ag++)
    if (agents[ag].isAlive)
      agents[ag].plan = agents[ag].move = agents[ag].migr = 0;
}
void testAndsetAgentplan (int id)
{
  I (id < nbmaxAg);
  I (!agents[id].plan);
  agents[id].plan = 1;
}
void testAndsetAgentmove (int id)
{
  I (id < nbmaxAg);
  I (!agents[id].move);
  agents[id].move = 1;
}
void testAndsetMAgent (int id)
{
  I (id < nbmaxAg);
  I (!agents[id].migr);
  agents[id].migr = 1;  // migr to deal
}
void testAndclearMAgent (int id)
{
  I (id < nbmaxAg);
  I (agents[id].migr);
  agents[id].migr = 0;  // migr is dealt
}
void testAndclearAgents (void)
{
  int ag;
  for (ag=0 ; ag<nbmaxAg ; ag++)
    if (agents[ag].isAlive){
      I (agents[ag].plan);
      I (agents[ag].move);
      I (!agents[ag].migr);  // assume migr has been dealt
      agents[ag].plan = agents[ag].move = 0;  // clear agent
    }
}

void testAndclearCAgents (stack_t *s)
{
  int i;
  for (i=0 ; i<depthStack(s) ; i++)
    I (((agentinfo_t*)peekstructp(s, i, sizeof(agentinfo_t)))->created);
}
#endif

// -------------------------  rand  --------------------------

// return a (reproducible) random number using current temporary seed
static int agRand (void)
{
  // the cast is against the 2.1.1 glibc bug (which may return negative numbers)
  return (unsigned int)rand_r(&getTSD()->seed);
}

// return a (reproducible) random number in 0..max-1 using current temporary seed
static int agRandBound (int max)
{
  return agRand() % max;
}

// return a (reproducible) random number
int pmtmRand (void)
{
  return agRand();
}

// get a (reproducible) random and valid direction for agent ID and put it in DIR
// Return 1 if successful, 0 if no free square was found
int getAgentRandomMove (int id, int *dir)
{
  int y, x, i, crtdir;
  I (agents[id].isAlive);
  lockRandomAgent ();
  I (nbRandomAg < nbAg);
  nbRandomAg++;
  unlockRandomAgent ();
  y = pmgiGetAgenty(id);
  x = pmgiGetAgentx(id);
  crtdir = agRandBound (_nbDirs);
  for (i=0 ; i<_nbDirs ; i++){
    crtdir = (crtdir + 1) % _nbDirs;
    if (pmtadCanMove (y, x, crtdir)){
      *dir = crtdir;
      return 1;
    }
  }
  return 0;
}

int pmtmGetAgentRandomMove (int *dir)
{
  return getAgentRandomMove (getTSD()->crtAgentId, dir);
}

// --------------------------  general functions  ------------------

int isAlive (int id)
{
  I (id < nbmaxAg);
  return (agents[id].isAlive);
}

int isKilled (int id)
{
  I (id < nbmaxAg);
  return (agents[id].isKilled);
}

int getEmptyAgentNumber (void)
{
  return nbEmptyAg;
}

int getRandomAgentNumber (void)
{
  return nbRandomAg;
}

void clearRandomAgent (void)
{
  nbRandomAg = 0;
}

int pmgGetAgentNumber (void)
{
  return nbAg;
}

int getAgentNumbermax (void)
{
  return nbmaxAg;
}

int getAgentPriority (int id)
{
  I (id < nbmaxAg);
  I (agents[id].isAlive);
  return agents[id].priority;
}

int getAgentSlowness (int id)
{
  I (id < nbmaxAg);
  I (agents[id].isAlive);
  return agents[id].slowness;
}

pmpfAgent getAgentFunc (int id)
{
  I (id < nbmaxAg);
  I (agents[id].isAlive);
  return agents[id].func;
}

void *getAgentMemory (int id)
{
  I (id < nbmaxAg);
  I (agents[id].isAlive);
  return agents[id].memory;
}

int getAgentnewy (int id)
{
  I (id < nbmaxAg);
  I (agents[id].isAlive);
  return agents[id].newy;
}

int getAgentnewx (int id)
{
  I (id < nbmaxAg);
  I (agents[id].isAlive);
  return agents[id].newx;
}

int pmgiGetAgentActiontype (int id)
{
  I (id < nbmaxAg);
  I (agents[id].isAlive);
  return agents[id].actiontype;
}

int pmgaGetAgentActiontype (int y, int x)
{
  ICHECKCOORDS (y, x);
  return pmgiGetAgentActiontype (pmgaGetAgentId(y, x));
}

int pmgrGetAgentActiontype (int ry, int rx)
{
  return pmgaGetAgentActiontype (pmtraCoordy(ry), pmtraCoordx(rx));
}

int pmlmGetAgentActiontype (void)
{
  return pmgiGetAgentActiontype (getTSD()->crtAgentId);
}

void setAgentActiontype (int id, int type)
{
  I (id < nbmaxAg);
  I (agents[id].isAlive);
  agents[id].actiontype = type;
}

void setAgentAction (int id, pmAction_t a)
{
  I (id < nbmaxAg);
  I (agents[id].isAlive);
  agents[id].actiontype = a.type;
  agents[id].actionparam = a.param;
}

int pmlmGetAgentId (void)
{
  return getTSD()->crtAgentId;
}

int pmgiGetAgenty (int id)
{
  I (id < nbmaxAg);
  I (agents[id].isAlive);
  return agents[id].y;
}

int pmgmGetAgenty (void)
{
  return pmgiGetAgenty (getTSD()->crtAgentId);
}

int pmgiGetAgentx (int id)
{
  I (id < nbmaxAg);
  I (agents[id].isAlive);
  return agents[id].x;
}

int pmgmGetAgentx (void)
{
  return pmgiGetAgentx (getTSD()->crtAgentId);
}

int pmgiGetAgentMaxload (int id)
{
  I (id < nbmaxAg);
  I (agents[id].isAlive);
  return agents[id].maxload;
}

int pmgaGetAgentMaxload (int y, int x)
{
  ICHECKCOORDS (y, x);
  return pmgiGetAgentMaxload (pmgaGetAgentId(y, x));
}

int pmgrGetAgentMaxload (int ry, int rx)
{
  return pmgaGetAgentMaxload (pmtraCoordy(ry), pmtraCoordx(rx));
}

int pmlmGetAgentMaxload (void)
{
  return pmgiGetAgentMaxload (getTSD()->crtAgentId);
}

int pmgiGetAgentLoad (int id)
{
  I (id < nbmaxAg);
  I (agents[id].isAlive);
  return agents[id].load;
}

int pmgaGetAgentLoad (int y, int x)
{
  ICHECKCOORDS (y, x);
  return pmgiGetAgentLoad (pmgaGetAgentId(y, x));
}

int pmgrGetAgentLoad (int ry, int rx)
{
  return pmgaGetAgentLoad (pmtraCoordy(ry), pmtraCoordx(rx));
}

int pmlmGetAgentLoad (void)
{
  return pmgiGetAgentLoad (getTSD()->crtAgentId);
}

// planify a move (fill the agent.nextpos and do the checks)
// doAgentMove function does it if it succeeds the conflicts
void planAgentMove (int id, int dir)
{
  int y, x, newy, newx;
  I (id < nbmaxAg);
  I (agents[id].isAlive);
  ICHECKDIR (dir);
  y = agents[id].y;
  x = agents[id].x;
  newy = normy (y + _diry[dir]);
  newx = normx (x + _dirx[dir]);
  // assume not wall on the square
  I (pmgaGetSquareType(newy, newx) != WALLSQ);
  // assume the square where it comes from is occupied
  I (pmgaGetSquareState(y, x) == OCCUP);
  // assume the square to move is not occupied (ag can't go one behind another)
  I (pmgaGetSquareState(newy, newx) == UNOCCUP);
  agents[id].newy = newy;
  agents[id].newx = newx;
}

void doAgentMove (int id)
{
  agent_t aginfo = agents[id];
  I (id < nbmaxAg);
  I (agents[id].isAlive);
  // assume it is really a move
  I ((aginfo.actiontype == MOVE) || (aginfo.actiontype == MOVE_CONFLICT)
     || (aginfo.actiontype == MOVERANDOM_MOVE) || (aginfo.actiontype == MOVERANDOM_MOVE_CONFLICT));
  // assume move from OCCUP to UNOCCUP
  I (pmgaGetSquareState(aginfo.y, aginfo.x) == OCCUP);
  I (pmgaGetSquareState(aginfo.newy, aginfo.newx) == UNOCCUP);
  setSquareAgentId (aginfo.newy, aginfo.newx, id);  // check for UNOCCUP, so this is done before the next two lines
  setSquareState (aginfo.y, aginfo.x, UNOCCUP);
  setSquareState (aginfo.newy, aginfo.newx, OCCUP);
  agents[id].y = aginfo.newy;
  agents[id].x = aginfo.newx;
}

void doAgentTake (int id)
{
  int srctype, srcid;  // info about the source
  agent_t aginfo = agents[id];
  I (id < nbmaxAg);
  I (agents[id].isAlive);
  // assume it is really a take
  I (aginfo.actiontype == TAKE);
  // assume it is on a source
  I (pmgaGetSquareType(aginfo.y, aginfo.x) == RESSQ);
  // decrement the source load
  srctype = getSquareSourceType (aginfo.y, aginfo.x);
  srcid = getSquareSourceId (aginfo.y, aginfo.x);
  changeResourceLoad (srctype, srcid, -aginfo.actionparam);
  // increment if needed empty agents
  I (aginfo.actionparam != 0);  // to easy computing nbEmptyMAg
  if (agents[id].load == 0){
    lockEmptyAgent ();
    I (nbEmptyAg > 0);
    nbEmptyAg--;
    unlockEmptyAgent ();
  }
  // and increment the agent load
  I (aginfo.load + aginfo.actionparam <= aginfo.maxload);
  agents[id].load += aginfo.actionparam;
}

void doAgentDrop (int id)
{
  int srctype, srcid;  // info about the source
  agent_t aginfo = agents[id];
  I (id < nbmaxAg);
  I (agents[id].isAlive);
  // assume it is really a drop
  I (aginfo.actiontype == DROP);
  // assume it is on a source
  I (pmgaGetSquareType(aginfo.y, aginfo.x) == RESSQ);
  // increment the source load (unlimited source load when drop)
  srctype = getSquareSourceType (aginfo.y, aginfo.x);
  srcid = getSquareSourceId (aginfo.y, aginfo.x);
  changeResourceLoad (srctype, srcid, aginfo.actionparam);
  // decrement if needed empty agents
  I (aginfo.actionparam != 0);  // to easy computing nbEmptyMAg
  if (agents[id].load == aginfo.actionparam){  // if drop all its load
    lockEmptyAgent ();
    I (nbEmptyAg < nbAg);
    nbEmptyAg++;
    unlockEmptyAgent ();
  }
  // and decrement the agent load
  I (aginfo.load >= aginfo.actionparam);
  agents[id].load -= aginfo.actionparam;
}

void initAgent (void)
{
  nbAg = nbmaxAg = 0;
  nbAllocAg = 0;
  agents = NULL;
  nbEmptyAg = 0;
  initStack (&deadAgents, INCDEAD);
}

void initAgentPar (void)
{
  initStack (&getTSD()->createdAgents, INCLOCALCA);
}

// put statically an agent in the beginning of simulation or loadSystem
static void addAgent (int y, int x, int slowness, int prio, int load, int maxl,
		      void *mem, int memsize, pmpfAgent func)
{
  ICHECKCOORDS (y, x);
  I ((pmgaGetSquareType(y, x) != WALLSQ) && (pmgaGetSquareState(y, x) == UNOCCUP));
  I ((prio == NORM_PR) || (prio == HIGH_PR));  // actually, only two priorities
  I (slowness != 0);  // the slowness is a strictly positive number
  if (nbAg == nbAllocAg){
    agent_t *tmp;
    tmp = realloc (agents, (nbAllocAg + INCAG) * sizeof(agent_t));
    if (tmp == NULL)
      abcerror ("Cannot reallocate %d bytes in addAgent.  Aborting\n", (nbAllocAg + INCAG) * sizeof(agent_t));
    agents = tmp;
    nbAllocAg += INCAG;
  }
  agents[nbAg].y = y;
  agents[nbAg].x = x;
  agents[nbAg].slowness = slowness;
  agents[nbAg].priority = prio;
  agents[nbAg].load = load;
  agents[nbAg].maxload = maxl;
  agents[nbAg].actiontype = CREATED;
  if (memsize == 0)
    agents[nbAg].memory = NULL;
  else{
    agents[nbAg].memory = malloc (memsize);
    if (agents[nbAg].memory == NULL){
      printf ("Cannot allocate %d bytes in pmgaCreateAgent.  Aborting.\n", memsize);
      exit (1);
    }
    memcpy (agents[nbAg].memory, mem, memsize);
  }
  agents[nbAg].func = func;
  agents[nbAg].isAlive = 1;
  agents[nbAg].isKilled = 0;
  X (agents[nbAg].plan = agents[nbAg].move = agents[nbAg].migr = 0);
  setSquareAgentId (y, x, nbAg);  // ckeck that the square is unoccup...
  setSquareState (y, x, OCCUP);  // ...so this line has to be executed after
  nbAg++;
  nbmaxAg++;
  if (load == 0)
    nbEmptyAg++;
}

// add an agent in agents structure during the simulation (preemptive function)
static void addParallelAgent (int y, int x, int slowness, int prio, int maxl,
			      void *mem, pmpfAgent func)
{
  int agindex = -1;  // the index in the agents struct where the new agent will be put in
  ICHECKCOORDS (y, x);
  I ((pmgaGetSquareType(y, x) != WALLSQ) && (pmgaGetSquareState(y, x) == UNOCCUP));
  I ((prio == NORM_PR) || (prio == HIGH_PR));  // actually, only two priorities
  I (slowness != 0);  // the slowness is a strictly positive number
  // try to take a deadAgent
  lockDAgent ();
  if (depthStack(&deadAgents) != 0){
    agindex = popint (&deadAgents);
    I (!agents[agindex].isAlive);  // should be free element
  }
  unlockDAgent ();
  if (agindex == -1){  // need to create a new structure in agents
    lockAgent ();
    I (nbAg == nbmaxAg);  // there are no holes in the agents structure
    if (nbAg == nbAllocAg){
      agent_t *tmp;
      tmp = realloc (agents, (nbAllocAg + INCAG) * sizeof(agent_t));
      if (tmp == NULL)
	abcerror ("Cannot reallocate %d bytes in addParallelAgent.  Aborting\n", (nbAllocAg + INCAG) * sizeof(agent_t));
      agents = tmp;
      nbAllocAg += INCAG;
    }
    agindex = nbAg;
    unlockAgent ();
  }
  I (agindex != -1);
  // fill the necessary fields
  agents[agindex].y = y;
  agents[agindex].x = x;
  agents[agindex].slowness = slowness;
  agents[agindex].priority = prio;
  agents[agindex].load = 0;
  agents[agindex].maxload = maxl;
  agents[agindex].actiontype = CREATED;
  agents[agindex].memory = mem;
  agents[agindex].func = func;
  agents[agindex].isAlive = 1;
  agents[agindex].isKilled = 0;
  X (agents[agindex].plan = 1; agents[agindex].move = agents[agindex].migr = 0);  // assume it was planified
  setSquareAgentId (y, x, agindex);  // ckeck that the square is unoccup...
  setSquareState (y, x, OCCUP);  // ...so this line has to be executed after
  locknbAgent ();
  I (nbmaxAg >= nbAg);
  nbAg++;
  if (nbAg > nbmaxAg)
    nbmaxAg = nbAg;  // or nbmaxAg++
  unlocknbAgent ();
  lockEmptyAgent ();
  nbEmptyAg++;
  unlockEmptyAgent ();
  VL (("Agent with id %d created in (%d,%d)\n", agindex, y, x));
}

void putAgentAll (int y, int x, int slowness, int prio, int load, int maxl,
		  void *mem, int memsize, pmpfAgent func)
{
  ICHECKCOORDS (y, x);
  addAgent (y, x, slowness, prio, load, maxl, mem, memsize, func);
}

void pmPutAgent (int y, int x, int slowness, int prio, int maxl,
		 void *mem, int memsize, void *func)
{
  ICHECKCOORDS (y, x);
  putAgentAll (y, x, slowness, prio, 0, maxl, mem, memsize, (pmpfAgent)func);
}

void pmPutRandomAgents (int nb, int slowness, int prio, int maxl,
			void *mem, int memsize, void *func)
{
  int y, x;
  // there are sufficiently free squares for all the agents
  I (nb + getObstNumber() + nbAg <= _dimy * _dimx);  // agent + obst <= y*x
  while (nb-- > 0){
    do{
      y = globrand(_dimy);
      x = globrand(_dimx);
    }while ((pmgaGetSquareType(y, x) == WALLSQ) || (pmgaGetSquareState(y, x) == OCCUP));
    pmPutAgent (y, x, slowness, prio, maxl, mem, memsize, (pmpfAgent)func);
  }
}

// create an agent.  This function is preemptive (called by all the threads)
// try to create an agent in (y,x) with parameters slowness, prio etc.
// put the appropriate structure in local or global stack
void pmgaCreateAgent (int y, int x, int slowness, int prio, int maxl,
		      void *mem, int memsize, pmpfAgent func)
{
  agentinfo_t ag;
  ICHECKCOORDS (y, x);
  I ((pmgaGetSquareType(y, x) != WALLSQ) && (pmgaGetSquareState(y, x) == UNOCCUP));
  I ((prio == NORM_PR) || (prio == HIGH_PR));  // actually, only two priorities
  I (slowness != 0);  // the slowness is a strictly positive number
  // create structure
  if (memsize == 0)
    ag.memory = NULL;
  else{
    ag.memory = malloc (memsize);
    if (ag.memory == NULL){
      printf ("Cannot allocate %d bytes in pmgaCreateAgent.  Aborting.\n", memsize);
      exit (1);
    }
    memcpy (ag.memory, mem, memsize);
  }
  ag.y = y;
  ag.x = x;
  ag.slowness = slowness;
  ag.priority = prio;
  ag.maxload = maxl;
  ag.func = func;
  ag.parentid = getTSD()->id;
  ag.created = 0;  // NB: this is done for global stack agents too
  // is the agent created locally or not?
  if (isInMyDomain(y, x)){
    int agindex;
    // add the created agent to the local stack
    agindex = pushstruct (&getTSD()->createdAgents, &ag, sizeof(agentinfo_t));
    // add a new position of a created agent to the environment
    addNewPosCA (y, x, agindex);
  }else{
    // add the created agent to the global stack.  It will be put in a local stack later
    lockCAgent ();
    pushstruct (&_createdAgents, &ag, sizeof(ag));
    unlockCAgent ();
  }
  VL (("Agent creation planified in (%d,%d)\n", y, x));
}

void pmlrCreateAgent (int ry, int rx, int slowness, int prio, int maxl,
		      void *mem, int memsize, pmpfAgent func)
{
  pmgaCreateAgent (pmtraCoordy(ry), pmtraCoordx(rx),
		   slowness, prio, maxl, mem, memsize, func);
}

void pmldCreateAgent (int dir, int slowness, int prio, int maxl,
		      void *mem, int memsize, pmpfAgent func)
{
  ICHECKDIR (dir);
  pmlrCreateAgent (_diry[dir], _dirx[dir], slowness, prio, maxl,
		   mem, memsize, func);
}

// create, during the simulation, an agent at specified distance from creator
// check clockwise the circle around the agent, starting from a random position
// this implementation is wrongly not equitable if there are obstacles: squares
//   coming after N obstacles have N times higher chances to be chosen than
//   squares after no obstacle
void pmlCreateAgentDist (int d, int slowness, int prio, int maxl,
			 void *mem, int memsize, pmpfAgent func)
{
  int y, x, i, j;
  int start = pmtmRand () % (4*d);
  tsd_t *ptsd = getTSD();
  int crty = ptsd->crtAgenty, crtx = ptsd->crtAgentx;
  for (j=start ; j<4*d+start ; j++){
    // find the square
    i = j % (4*d);
    if (i < d){  // top -> right
      y = (crty - d) + i;
      x = crtx + i;
    }else if (i < 2*d){  // right -> bottom
      y = crty + (i - d);
      x = (crtx + d) - (i - d);
    }else if (i < 3*d){  // bottom -> left
      y = (crty + d) - (i - 2*d);
      x = crtx - (i - 2*d);
    }else{  // left -> top
      y = crty - (i - 3*d);
      x = (crtx - d) + (i - 3*d);
    }
    // check if possible to create agent in the square
    if (_envType == EDGES){  // edges
      if ((y >= 0) && (y < _dimy) && (x >= 0) && (x < _dimx)  // inside env.
	  && pmtaCanMove (y, x)){
	pmgaCreateAgent (y, x, slowness, prio, maxl, mem, memsize, func);
	return;
      }
    }else{  // torus
      y = normy (y);
      x = normx (x);
      if (pmtaCanMove (y, x)){
	pmgaCreateAgent (y, x, slowness, prio, maxl, mem, memsize, func);
	return;
      }
    }
  }
  printf ("Agent creation refused, no free square at distance %d.  Continuing.\n", d);
}

// create, during the simulation, an agent anywhere in the world
void pmlCreateAgentAnywhere (int slowness, int prio, int maxl,
			     void *mem, int memsize, pmpfAgent func)
{
  int y, x;
  if (getObstNumber() + nbAg == _dimy * _dimx){
    printf ("Agent creation refused, no free square in the environment.  Continuing.\n");
    return;
  }
  do{
    y = agRandBound (_dimy);
    x = agRandBound (_dimx);
  }while (pmtaCanMove(y, x));
  pmgaCreateAgent (y, x, slowness, prio, maxl, mem, memsize, func);
}

// this is a created agent which will be created (either alone or winner in a conflict)
// copy the agent with the appropriate localCreatedAgent index in agents struct
void processWinnerCA (int agindex)
{
  agentinfo_t *ag = peekstructp (&getTSD()->createdAgents, agindex, sizeof(agentinfo_t));
  I (!ag->created);
  ag->created = 1;  // it was processed
  addParallelAgent (ag->y, ag->x, ag->slowness, ag->priority, ag->maxload,
		    ag->memory, ag->func);
}

// this is an agent which has lost in a conflict, so it is lost
// (at this version of the library)
// Print a message about the lost agent
void processLooserCA (int agindex)
{
  agentinfo_t *ag = peekstructp (&getTSD()->createdAgents, agindex, sizeof(agentinfo_t));
  I (!ag->created);
  ag->created = 1;  // it was processed
  if (ag->memory)
    free (ag->memory);
  printf ("Agent creation in (%d,%d) refused\n", ag->y, ag->x);
}

// mark an agent as dead - reentrant function
void markDeadAgent (int id)
{
  I (id <= nbmaxAg);
  I (agents[id].isKilled);
  // remove it from the environment
  I (pmgaGetSquareState(agents[id].y, agents[id].x) == OCCUP);
  setSquareState (agents[id].y, agents[id].x, UNOCCUP);
  // update misc variables
  if (agents[id].load == 0){
    lockEmptyAgent ();
    I (nbEmptyAg > 0);
    nbEmptyAg--;
    unlockEmptyAgent ();
  }
  locknbAgent ();
  I (nbAg > 0);
  nbAg--;
  unlocknbAgent ();
  // mark the agent as dead
  I (agents[id].isAlive);
  agents[id].isAlive = 0;
  X (agents[id].isKilled = 0);
  // free its memory if necessary
  if (agents[id].memory)
    free (agents[id].memory);
  // add the agent to the deadAgents stack
  lockDAgent ();
  pushint (&deadAgents, id);
  unlockDAgent ();
  VL (("Agent %d killed by thread %d\n", id, getTSD()->id));
}

// mark an agent to be killed
void pmgiKillAgent (int id)
{
  I (id <= nbmaxAg);
  I (agents[id].isAlive);
  I (! agents[id].isKilled);
  agents[id].isKilled = 1;  // here, race condition when it is killed by 2 threads in the same time
}

void pmgaKillAgent (int y, int x)
{
  ICHECKCOORDS (y, x);
  pmgiKillAgent (pmgaGetAgentId(y, x));
}

void pmlrKillAgent (int ry, int rx)
{
  pmgaKillAgent (pmtraCoordy(ry), pmtraCoordx(rx));
}

void pmldKillAgent (int dir)
{
  ICHECKDIR (dir);
  pmlrKillAgent (_diry[dir], _dirx[dir]);
}

void pmlmKillAgent (void)
{
  pmgiKillAgent (getTSD()->crtAgentId);
}

void freeAgent (void)
{
  clearStack (&deadAgents);
  endStack (&deadAgents);
  if (agents){
    int i;
    for (i=0 ; i<nbmaxAg ; i++)
      if (agents[i].isAlive && agents[i].memory)
	free (agents[i].memory);
    free (agents);
  }
}
