// This file, along with libagents.a library, have to be used by applications

// constants

// these error codes must be negative numbers, because probably (to verify)
// there are functions which return useful data (positive numbers) or error
#define OK      0
#define ERR     -1
#define FILEERR -2      // file error
#define FUNCERR -3      // no function in agfuncs or l2pfuncs sets
                        // corresponding to id found in config file.  Save also

// square types
#define FREESQ  0
#define RESSQ   1
#define WALLSQ  2

// square states
#define UNOCCUP 0
#define OCCUP   1

// env types
#define TORUS   0
#define EDGES   1

// agent priorities
#define NORM_PR 0
#define HIGH_PR 1
// --- end init with 0

// vision types
#define DYNAMIC 0
#define STATIC  1

// propagation methods
// fixed domain decomp, recursive breadth-first for domain propagation
#define BREADTH_BREADTH_DOMAIN     0  // breadth-first for frontier propagation
#define BREADTH_DEPTH_DOMAIN       1  // depth-first for frontier propagation
#define BREADTH_ITERATIVE_DOMAIN   2  // iterative for frontier propagation
// recursive depth-first, fixed domain decomp
#define DEPTH_BREADTH_DOMAIN       3  // breadth-first for frontier propagation
#define DEPTH_DEPTH_DOMAIN         4  // depth-first for frontier propagation
#define DEPTH_ITERATIVE_DOMAIN     5  // iterative for frontier propagation
// iterative variable potential, fixed domain decomp
#define ITERATIVE_BREADTH_DOMAIN   6  // breadth-first for frontier propagation
#define ITERATIVE_DEPTH_DOMAIN     7  // depth-first for frontier propagation
#define ITERATIVE_ITERATIVE_DOMAIN 8  // iterative for frontier propagation
// private environments
#define BREADTH_PRIVATE            9  // recursive breadth-first
#define ITERATIVE_PRIVATE          10 // iterative variable potential

// runfunction return values
#define CONT    0
#define STOP    1
// ---

// NB: The rectangle directions have to be 0..3 (used by setMouvement, I)
//   the others 4..7
//   the order of all these constants matters for pmtDir2y() and pmtDir2x()
#define UP      0
#define DOWN    1
#define LEFT    2
#define RIGHT   3
#define UPL     4
#define UPR     5
#define DOL     6
#define DOR     7

// types of action
// the actions set by the application
#define STAY    0
#define MOVE    1
#define TAKE    2  // only if it is a source in the same square
#define DROP    3  // only if it is a source in the same square
#define MOVERANDOM  4  // random movement done by the library
#define CREATED 5  // the first life cycle of the agent
// the actions set by the system (the application may only read them)
#define PASS    6  // when the function is not called because of the slowness.  Used also by the system at do_moves
#define STAY_CONFLICT 7                // loser in a conflict
#define STAY_CA_CONFLICT 8             // loser in conflict with created agents
#define MOVE_CONFLICT 9                // it won the conflict
#define MOVERANDOM_STAY 10             // no square to move
#define MOVERANDOM_MOVE 11             // plan a move for random
#define MOVERANDOM_MOVE_CONFLICT 12    // winner in a conflict
#define MOVERANDOM_STAY_CONFLICT 13    // loser in a conflict
#define MOVERANDOM_STAY_CA_CONFLICT 14 // loser in conflict with created agents


// -----------------------------------------------------------------


// types

// the pmAction_t structure used by the agents
typedef struct{
  int type;   // STAY, MOVE etc. (which may be defined by the user)
  int param;  // dir (if MOVE), value (if DROP, TAKE) or other useful value
}pmAction_t;

typedef pmAction_t (*pmpfAgent) (void *);           // behaviour func of agent - in: memory
typedef int (*pmpfSource)     (int, int, int);      // load, initload, initpot
    // load->pot function (potential radius is equal to pot)
    // checked condition: func (initload, initload, initpot) == initpot
typedef int (*pmpfCallback)   (void);               // return CONT or STOP


// -----------------------------------------------------------------


// --------  system functions
void pmInitSystem         (unsigned int);       // global seed

void pmRunSystem          (pmpfCallback);
    // pmpfCallback is the function called at the end of every cycle

void pmEndSystem          (void);


// ---------  environment functions
void pmCreateEnv          (int, int, int);      // dimY, dimX, type
    // type: torus or edge. If edge, obstacles are added automatically at edges
    // y and x range from 0 to dim-1

void pmPutObstacle        (int, int);           // y, x
void pmPutRandomObstacles (int);                // nb

void pmPutAgent           (int, int, int, int, int, void*, int, void*);  // y, x, slowness, priority, maxload, memory, memsize
void pmPutRandomAgents    (int, int, int, int, void*, int, void*);  // nb, slowness, priority, maxload, memory
    // func is the movement function.  See optimizations
    // priority: high, normal

void pmPutResource        (int, int, int, int, int, pmpfSource);  // y, x, type, init_load, init_pot
void pmPutRandomResources (int, int, int, int, pmpfSource);  // nb, type, init_load, init_pot
void pmSetPropagationMethod (int);              // method (see #define above)
void pmSetResRefreshPeriod(int);                // 0-init only, n-each n cycles
    // pmpfSource is the load -> pot(rp) function.  See optimizations
    // every time we add a source type, an int (for its potential) is
    //   added to the env structure
    // getResourceLastLoad assumes potentials are updated when pmLoadSystem
    // (you have to update potentials before pmSaveSystem, else lastload is garbage)
void pmSetVisionRadius    (int);                // 0-no vision, n-radius
void pmSetVisionType      (int);                // STATIC or DYNAMIC
void pmSetVisionSymmetry  (int);                // !=0 means symmetry


// ---------------  I/O functions
// checkpoint functions
void pmSetParallelSaving  (int);                // 0 or 1
int  pmSaveSystem         (char *, pmpfAgent *, pmpfSource *);  // baseFileName, funcs
int  pmLoadSystem         (char *, pmpfAgent *, pmpfSource *);  // baseFileName, funcs
    // the func pointers are sets of all respective functions, ended with NULL
    //   they must be unchanged for pmSaveSystem and pmLoadSystem of same files

// partial data functions
int  pmInitChanges        (char *);             // baseFileName
int  pmSaveChanges        (void);               //
int  pmInitStat           (char *);             // filename for statistic file
void pmSaveStat           (void);               // write info in statistic file


// ----------  general functions  -----------------------

void pmSetDirsNo          (int);      // n=4 or 8: set the number of dirs used
void pmSetThreadsNo       (int);      // set the number of all the threads used
                                      // this creates PARAM-1 threads


// -------------  useful functions in funcag  ------------------
// ----  local functions can be used ONLY in funcag (implicit id, y and x)  --

// y, x, slowness, priority, maxload, memory, memsize
void pmgaCreateAgent      (int, int, int, int, int, void*, int, pmpfAgent);
void pmlrCreateAgent      (int, int, int, int, int, void*, int, pmpfAgent);
void pmldCreateAgent      (int, int, int, int, void*, int, pmpfAgent);
void pmlCreateAgentDist   (int, int, int, int, void*, int, pmpfAgent);
void pmlCreateAgentAnywhere(int, int, int, void*, int, pmpfAgent);

void pmgiKillAgent        (int);                 // id
void pmgaKillAgent        (int, int);            // y, x
void pmlrKillAgent        (int, int);            // ry, rx
void pmldKillAgent        (int);                 // dir
void pmlmKillAgent        (void);

int  pmgaGetAgentId       (int, int);            // y, x
int  pmlrGetAgentId       (int, int);            // ry, rx
int  pmldGetAgentId       (int);                 // dir
int  pmlmGetAgentId       (void);

int  pmgiGetAgenty        (int);                 // id
int  pmgmGetAgenty        (void);
int  pmgiGetAgentx        (int);                 // id
int  pmgmGetAgentx        (void);

int  pmgiGetAgentActiontype (int);               // id
int  pmgaGetAgentActiontype (int, int);          // y, x
int  pmgrGetAgentActiontype (int, int);          // ry, rx
int  pmlmGetAgentActiontype (void);
int  pmgiGetAgentMaxload  (int);                 // id
int  pmgaGetAgentMaxload  (int, int);            // y, x
int  pmgrGetAgentMaxload  (int, int);            // ry, rx
int  pmlmGetAgentMaxload  (void);
int  pmgiGetAgentLoad     (int);                 // id
int  pmgaGetAgentLoad     (int, int);            // y, x
int  pmgrGetAgentLoad     (int, int);            // ry, rx
int  pmlmGetAgentLoad     (void);

int  pmgaGetSquareType    (int, int);            // y, x
int  pmlrGetSquareType    (int, int);            // ry, rx
int  pmldGetSquareType    (int);                 // dir
int  pmlmGetSquareType    (void);
int  pmgaGetSquareState   (int, int);            // y, x
int  pmlrGetSquareState   (int, int);            // ry, rx
int  pmldGetSquareState   (int);                 // dir
int  pmlmGetSquareState   (void);

int  pmgGetResTotalLoad   (int);                 // type
int  pmgaGetSquareResType (int, int);            // y, x
int  pmlrGetSquareResType (int, int);            // ry, rx
int  pmldGetSquareResType (int);                 // dir
int  pmlmGetSquareResType (void);
int  pmgaGetSquareResLoad (int, int);            // y, x
int  pmlrGetSquareResLoad (int, int);            // ry, rx
int  pmldGetSquareResLoad (int);                 // dir
int  pmlmGetSquareResLoad (void);
int  pmgaGetSquarePotential(int, int, int);      // source type, y, x
int  pmlrGetSquarePotential(int, int, int);      // source type, ry, rx
int  pmldGetSquarePotential(int, int);           // source type, dir

int  pmgaaIsVisibleSquare (int, int, int, int);  // y1, x1, y2, x2
int  pmgiaIsVisibleSquare (int, int, int);       // agent, y, x
int  pmgmaIsVisibleSquare (int, int);            // y, x
int  pmlrIsVisibleSquare  (int, int);            // ry, rx
int  pmldIsVisibleSquare  (int);                 // dir
//TODO: pmgir               (id,ry,rx)
//TODO: pmgmi               (id)

int  pmgGetEnvDimy        (void);
int  pmgGetEnvDimx        (void);
int  pmgGetCycle          (void);                // return the current cycle
int  pmgGetAgentNumber    (void);                // return the number of agents

int  pmtarCoordy          (int);                 // absolute->relative y
int  pmtarCoordx          (int);                 // absolute->relative x
int  pmtraCoordy          (int);                 // relative->absolute y
int  pmtraCoordx          (int);                 // relative->absolute x
int  pmtDir2y             (int);                 // dir -> relative y
int  pmtDir2x             (int);                 // dir -> relative x
// return 1 if an agent may move in (y,x), else 0
int  pmtaCanMove          (int, int);            // y, x
// return 1 if an agent found in (y,x) may move in direction DIR, else 0
int  pmtadCanMove         (int, int, int);       // y, x, dir
int  pmtmrCanMove         (int, int);            // ry, rx
int  pmtdCanMove          (int);                 // dir
int  pmtmGetAgentRandomMove(int *);              // dir, return 0 if can't move

// unique random function to be used in the application for reproducibility
int  pmtmRand             (void);                // return reproducible random
