// This source file deals with gradient sources

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "agents.h"
#include "intern.h"
#include "lowinterface.h"
#include "nana.h"

#define INCGS	 100   // allocates by INCGS the gradient sources
#define INCTYPES 5     // allocates by INCTYPES the types of sources
#define DELIM    60000 // a number > _dimy in any simulation (used as delimiter)
#define MAXPOT   32    // the max pot of any of the sources (the queue is growing anyway)

static int propMethod = BREADTH_BREADTH_DOMAIN;

static int gradsrcUpdate;  // the period of gradient update: 0-init, n-every n cycles

static int ***globalPots;  // 3D: thread, y, x - used in privateDomains()

static int nbTotalRes;          // total number of src of all the types
static int initResLoad;  // total initial load (useful for statistics) - it's only one ore !

static int *frChanged[2];  // 2D matrix => avoid a sync

// a gradient source is uniquely identified by type and number
typedef struct{
  pmpfSource func;
  int y, x;
  int type;
  int load;
  int initload;         // the initial load
  int initpot;          // the initial potential, which allows the same function for different initLoad and initPotential
  int pot;		// optional field (= load2rp(load)), but it speeds-up the simulation.  OTOH, it is required when saving (used instead lastload)
}gradsrc_t;

typedef struct{
  int usertype;     // the type as declared by the user
  int totalload;    // total load of these sources
  int unvisitednb;  // number of unvisited sources
  int emptynb;      // number of empty sources
  int **pots;       // the potential of all the squares of this type
  stack_t src;      // vector of gradsrc_t, all the sources of this type
}srctype_t;

static stack_t gss;  // vector of srctype_t

// get the maximum field of a source
// used for the queue allocation
// this formula is proven in my paper
static int maxfield (int srcpot)
{
  // NB: _nbDirs/8 is needed to add a square if 8 dirs
  return _nbDirs/4 * (2*srcpot*srcpot - 2*srcpot + 1) + _nbDirs/8;
}

// get a pointer to the data of type TYPE
static srctype_t *gettype (int type)
{
  I (type != -1);
  return (srctype_t*)peekstructp (&gss, type, sizeof(srctype_t));
}

// get a pointer to the source number IND
static gradsrc_t *getsrc (srctype_t *ags, int ind)
{
  return (gradsrc_t*)peekstructp (&ags->src, ind, sizeof (gradsrc_t));
}

// get the internal type of the usertype given as parameter
// the potential on the environment is written as usertype
// return -1 if usertype not found
static int user2type (int usertype)
{
  int type;
  for (type=0 ; type<depthStack(&gss) ; type++)
    if (gettype(type)->usertype == usertype)
      return type;
  return -1;  // not found
}

// get the usertype of the type given as parameter
int type2user (int type)
{
  I (type < depthStack(&gss));
  return gettype(type)->usertype;
}

// NB: this function is called in the beginning, so cannot use here _nbThreads
void initResource (void)
{
  I (_dimy < DELIM);  // DELIM is used as a special case
  gradsrcUpdate = 1;
  initResLoad = 0;
  nbTotalRes = 0;
  initStack (&gss, INCTYPES);
}

// this function is called in the beginning of pmRunSystem and just before
// the thread creation, so here we can use safely _nbThreads
void initResource2 (void)
{
  globalPots = calloc (_nbThreads, sizeof (int**));  // NB: calloc sets to NULL
  frChanged[0] = calloc(depthStack(&gss), sizeof(int));  // NB: set to NULL
  frChanged[1] = malloc(depthStack(&gss) * sizeof(int));
  if ((frChanged[0] == NULL) || (frChanged[1] == NULL))
    abcerror ("Cannot allocate %d bytes twice in initResource2.  Aborting.\n", depthStack(&gss) * sizeof(int));
}

void initResourcePar (void)
{
  tsd_t *ptsd = getTSD();
  initQueue (&ptsd->srcQueue, 2 * maxfield(MAXPOT));  // 2 ints/square
  ptsd->nfrontier = ptsd->efrontier = ptsd->sfrontier = ptsd->wfrontier = NULL;
  // dynamic allocation of frontChanged
  ptsd->frontChanged = malloc (depthStack(&gss) * sizeof(int));
  if (ptsd->frontChanged == NULL)
    abcerror ("Cannot allocate %d bytes in initResourcePar.  Aborting.\n", depthStack(&gss) * sizeof(int));
}

int getUnvisitedResNumber (int type)
{
  srctype_t *ags = gettype(type);
  return ags->unvisitednb;
}

int getEmptyResNumber (int type)
{
  srctype_t *ags = gettype(type);
  return ags->emptynb;
}

int getResTotalLoad (int type)
{
  srctype_t *ags = gettype(type);
  return ags->totalload;
}

int pmgGetResTotalLoad (int usertype)
{
  int type = user2type (usertype);
  I (type != -1);
  return getResTotalLoad (type);
}

int getInitResLoad (void)
{
  return initResLoad;
}

int getResourceLoad (int type, int nb)
{
  srctype_t *ags = gettype(type);
  return getsrc(ags, nb)->load;
}

void changeResourceLoad (int type, int nb, int ldchange)
{
  srctype_t *ags = gettype(type);
  gradsrc_t *src = getsrc(ags, nb);
  I (ldchange);    // to easy compute the emptyResLoad
  I (src->load + ldchange >= 0);
  if (src->load == 0){    // this will become diff. than 0
    lockEmptyResN (type);
    I (ags->emptynb > 0);
    ags->emptynb--;
    unlockEmptyResN (type);
  }
  // next: bug if its load grows and lowers...
  if (src->load == src->initload){  // it becomes visited
    lockUnvisResN (type);
    // I (ags->unvisitednb > 0);
    ags->unvisitednb--;
    unlockUnvisResN (type);
  }
  src->load += ldchange;
  if (src->load == 0){
    lockEmptyResN (type);
    ags->emptynb++;    // it becomes 0
    unlockEmptyResN (type);
  }
  lockTotalResL (type);
  I (ags->totalload + ldchange >= 0);
  ags->totalload += ldchange;
  unlockTotalResL (type);
}

int getResourceInitLoad (int type, int nb)
{
  srctype_t *ags = gettype(type);
  return getsrc(ags, nb)->initload;
}

int getResourcePot (int type, int nb)
{
  srctype_t *ags = gettype(type);
  return getsrc(ags, nb)->pot;
}

int getResourceInitPot (int type, int nb)
{
  srctype_t *ags = gettype(type);
  return getsrc(ags, nb)->initpot;
}

int getResourcey (int type, int nb)
{
  srctype_t *ags = gettype(type);
  return getsrc(ags, nb)->y;
}

int getResourcex (int type, int nb)
{
  srctype_t *ags = gettype(type);
  return getsrc(ags, nb)->x;
}

pmpfSource getResourceFunc (int type, int nb)
{
  srctype_t *ags = gettype(type);
  return getsrc(ags, nb)->func;
}

int getResourceNumber (int type)
{
  srctype_t *ags = gettype(type);
  return depthStack (&ags->src);
}

int getResourceTypesNumber (void)
{
  return depthStack(&gss);
}

int getTotalResourceNumber (void)
{
  return nbTotalRes;
}

// add and initialize a gradient source in the structure of gradient sources
static void addResource (int y, int x, int type, int load, int initload,
			 int pot, int initpot, pmpfSource func)
{
  gradsrc_t newsrc;
  int ind;
  srctype_t *ags = gettype(type);
  ICHECKCOORDS (y, x);
  // if func is not NULL, then for load==L0, pot==P0
  I (!func || func(initload, initload, initpot) == initpot);
  newsrc.y = y;
  newsrc.x = x;
  newsrc.type = type;
  newsrc.load = load;
  newsrc.initload = initload;
  newsrc.pot = pot;
  newsrc.initpot = initpot;
  newsrc.func = func;
  ind = pushstruct (&ags->src, &newsrc, sizeof(newsrc));
  ags->totalload += initload;
  initResLoad += initload;  // or set before simulation beginning
  if (load == 0)
    ags->emptynb++;
  if (load == initload)  // if come from initial, not from pmLoadSystem
    ags->unvisitednb++;
  putSquareGS (y, x, type, ind);
  nbTotalRes++;
}

// pot is also given as parameter
// NB: this is the only function which does not start with pm and uses usertype
void putUsertypeResource (int y, int x, int usertype, int load, int initload,
			  int pot, int initpot, pmpfSource func)
{
  int type = user2type(usertype);
  ICHECKCOORDS (y, x);
  // don't allow two src in the same square or in a wall square
  I (pmgaGetSquareType(y, x) == FREESQ);
  if (type == -1){  // it is the first time this type is used => initialize
    int i;
    srctype_t newtype;
    newtype.usertype = usertype;
    newtype.totalload = 0;
    newtype.unvisitednb = 0;
    newtype.emptynb = 0;
    initStack (&newtype.src, INCGS);
    // pots allocation
    newtype.pots = malloc (_dimy * sizeof(int*));
    if (newtype.pots == NULL)
      abcerror ("Cannot allocate %d bytes in putUsertypeResource.  Aborting\n", _dimy * sizeof(int*));
    newtype.pots[0] = malloc (_dimy * _dimx * sizeof(int));
    if (newtype.pots[0] == NULL)
      abcerror ("Cannot allocate %d bytes in putUsertypeResource.  Aborting\n", _dimy * _dimx * sizeof(int));
    for (i=1 ; i<_dimy ; i++)
      newtype.pots[i] = newtype.pots[0] + i*_dimx;
    type = pushstruct (&gss, &newtype, sizeof(newtype));
  }
  addResource (y, x, type, load, initload, pot, initpot, func);
}

// same as above, but without the pot parameter
void pmPutResource (int y, int x, int usertype, int initload, int initpot,
		    pmpfSource func)
{
  ICHECKCOORDS (y, x);
  putUsertypeResource (y, x, usertype, initload, initload, initpot, initpot, func);
}

void pmPutRandomResources (int nb, int usertype, int initload, int initpot,
			   pmpfSource func)
{
  int y, x;
  // there are sufficient free squares for all the sources
  I (nb + getObstNumber() + nbTotalRes <= _dimy * _dimx);  // agent + gsrc <= y*x
  while (nb-- > 0){
    do{
      y = globrand(_dimy);
      x = globrand(_dimx);
    }while (pmgaGetSquareType(y, x) != FREESQ);
    pmPutResource (y, x, usertype, initload, initpot, func);
  }
}

void freeResourcePar (void)
{
  tsd_t *ptsd = getTSD();
  endQueue (&ptsd->srcQueue);
  if (ptsd->nfrontier)
    free (ptsd->nfrontier);
  if (ptsd->sfrontier)
    free (ptsd->sfrontier);
  if (ptsd->wfrontier)
    free (ptsd->wfrontier);
  if (ptsd->efrontier)
    free (ptsd->efrontier);
  if (globalPots[ptsd->id]){
    free (globalPots[ptsd->id][0]);
    free (globalPots[ptsd->id]);
  }
  free (ptsd->frontChanged);
}

void freeResource (void)
{
  int type;
  for (type=0 ; type<depthStack(&gss) ; type++){
    srctype_t *ags = gettype(type);
    I (ags->pots[0]);  // I do not know exactly if this must be true
    free (ags->pots[0]);
    free (ags->pots);
    clearStack (&ags->src);  // do not check the size - needed for endStack
    endStack (&ags->src);
  }
  clearStack (&gss);  // do not check the size - needed for endStack
  endStack (&gss);
  free (frChanged[0]);
  free (frChanged[1]);
  free (globalPots);
}


// --------------------------  potential  ----------------------------


// return the potential of type-source type in square (y, x)
int pmgaGetSquarePotential (int usertype, int y, int x)
{
  ICHECKCOORDS (y, x);
  return gettype(user2type(usertype))->pots[y][x];
}

int pmlrGetSquarePotential (int usertype, int ry, int rx)
{
  return pmgaGetSquarePotential (usertype, pmtraCoordy(ry), pmtraCoordx(rx));
}

int pmldGetSquarePotential (int usertype, int dir)
{
  ICHECKDIR (dir);
  return pmlrGetSquarePotential (usertype, _diry[dir], _dirx[dir]);
}

void pmSetResRefreshPeriod (int upd)
{
  I (upd >= 0);
  gradsrcUpdate = upd;
}

int getResourcePeriod (void)
{
  return gradsrcUpdate;
}

// return 1 if need update potentials
// TODO: verify also if a source changes its load in order to update the pots
int needupdate (void)
{
  if (gradsrcUpdate == 0)
    return 0;
  return (getTSD()->cycle % gradsrcUpdate) == 0;
}

// clear the potential of the squares of its domain (parallel-called function)
// function common to all the methods of updateGradients
static void clearPotentials (void)
{
  int y, x;
  int type;
  tsd_t *ptsd = getTSD();
  srctype_t *ags;
  for (type=0 ; type<depthStack(&gss) ; type++){
    ags = gettype(type);
    for (y=ptsd->ytop ; y<ptsd->ybot ; y++)
      // some tests showed that the next for is faster than memset if dx<=128
      for (x=ptsd->xtop ; x<ptsd->xbot ; x++)
	ags->pots[y][x] = 0;
  }
}


/* ***************  miscellaneous *********************** */


// the difference with isInMyDomain is that this method is static, so inlined
static int isInMyDomain_ (int y, int x, int ytop, int xtop, int ybot, int xbot)
{
  ICHECKCOORDS (y, x);
  return ((ytop <= y) && (y < ybot) && (xtop <= x) && (x < xbot));
}

/* // return 1 if on environment frontier (0 or dim-1), else 0 */
/* // used to find out if norm needs to be used or not */
/* static int isOnEnvFrontier (int y, int x) */
/* { */
/*   ICHECKCOORDS (y, x); */
/*   return (y == 0) || (y == _dimy-1) || (x == 0) || (x == _dimx-1); */
/* } */

// return 1 if it is on the domain frontier, else 0
// NB: if torus, one point can be on env frontier and not on domain frontier
// used to find out if setChangedFrontier has to be called
// function common to all the methods of updateGradients
static int isOnDomainFrontier (int y, int x, int ytop, int xtop, int ybot, int xbot)
{
  int result;
  ICHECKCOORDS (y, x);
  result = (y == ytop) || (y == ybot-1) || (x == xtop) || (x == xbot-1);
/*   // next method not so sure... */
/*   // check against the four corner adjacent squares */
/*   result = (isInMyDomain_(normy(y-1), normx(x-1), ytop, xtop, ybot, xbot) */
/* 	    && isInMyDomain_(normy(y-1), normx(x+1), ytop, xtop, ybot, xbot) */
/* 	    && isInMyDomain_(normy(y+1), normx(x-1), ytop, xtop, ybot, xbot) */
/* 	    && isInMyDomain_(normy(y+1), normx(x+1), ytop, xtop, ybot, xbot)); */
  // VL (("isOnDomainFrontier: %d %d %d, %d %d %d, %d\n", y, ytop, ybot, x, xtop, xbot, result));
  return result;
/*   // it is difficult for me to find a correct method like below */
/*   // not on domain extremes => not on domain frontier */
/*   if ((ytop != y) && (y != ybot-1) && (xtop != x) && (x != xbot-1)) */
/*     return 0; */
/*   // on environment extremes <=> not on domain frontier */
/*   if (isOnEnvFrontier (y, x)) */
/*     return 0; */
/*   return 1; */
}


/* ***********  fixed domain decomposition parallel method  **************** */

// parallel method using boundary exchanges
// Pre-conditions: frontChanged must have already been set accordingly
//   no need of threads begin synchronized
// Post-conditions: the field is correctly propagated through all the env
//   the threads are synchronized at the end
// it is not useful to optimize the sync points by using only the neighbours, because, anyway, there are synchronization points during repropagations and the spare threads are forced to wait
static void boundaryExchg (void (*repropfunc)(int))
{
  int ind = 0;  // index to frChanged (used as bool)
  int type, i, max, loop;
  tsd_t *ptsd = getTSD();
  int ytop = ptsd->ytop, ybot = ptsd->ybot, xtop = ptsd->xtop, xbot = ptsd->xbot;
  do{
    loop = 0;
    // set frChanged
    for (type=0 ; type<depthStack(&gss) ; type++)
      if (ptsd->frontChanged[type]){
	lockfrch ();
	frChanged[ind][type] = 1;  // store in a global variable
	unlockfrch ();
      }
    syncThreads ();
    ind = 1 - ind;  // switch the index
    if (ptsd->id == 0)  // clear frChanged
      for (type=0 ; type<depthStack(&gss) ; type++)
	frChanged[ind][type] = 0;
    for (type=0 ; type<depthStack(&gss) ; type++)
      if (frChanged[1-ind][type]){
	int **pots = gettype(type)->pots;
	int update;  // used for all the frontiers (until a better solution)
	VL (("boundaryExchg needReprop: thread %d, usertype %d\n", ptsd->id, type2user(type)));
	loop = 1;
	// copy the frontiers
	// TODO: check here which frontiers need to be copied (neighb or not?) - use correct functions
	if (ytop != 0){  // copy N frontier
	  if (ptsd->nfrontier == NULL)  // the env. coordinates do not change
	    if ((ptsd->nfrontier = malloc ((xbot-xtop) * sizeof(int))) == NULL)
	      abcerror ("Cannot allocate %d bytes in boundaryExchg.  Aborting.\n", (xbot-xtop) * sizeof(int));
	  memcpy (ptsd->nfrontier, &pots[ytop-1][xtop], (xbot-xtop)*sizeof(int));
	}
	if (ybot != _dimy){  // copy S frontier
	  if (ptsd->sfrontier == NULL)  // the env. coordinates do not change
	    if ((ptsd->sfrontier = malloc ((xbot-xtop) * sizeof(int))) == NULL)
	      abcerror ("Cannot allocate %d bytes in boundaryExchg.  Aborting.\n", (xbot-xtop) * sizeof(int));
	  memcpy (ptsd->sfrontier, &pots[ybot][xtop], (xbot-xtop) * sizeof(int));
	}
	if (xtop != 0){  // copy W frontier
	  if (ptsd->wfrontier == NULL)  // the env. coordinates do not change
	    if ((ptsd->wfrontier = malloc ((ybot-ytop) * sizeof(int))) == NULL)
	      abcerror ("Cannot allocate %d bytes in boundaryExchg.  Aborting.\n", (ybot-ytop) * sizeof(int));
	  for (i=0 ; i<ybot-ytop ; i++)
	    ptsd->wfrontier[i] = pots[ytop+i][xtop-1];
	}
	if (xbot != _dimx){  // copy E frontier
	  if (ptsd->efrontier == NULL)  // the env. coordinates do not change
	    if ((ptsd->efrontier = malloc ((ybot-ytop) * sizeof(int))) == NULL)
	      abcerror ("Cannot allocate %d bytes in boundaryExchg.  Aborting.\n", (ybot-ytop) * sizeof(int));
	  for (i=0 ; i<ybot-ytop ; i++)
	    ptsd->efrontier[i] = pots[ytop+i][xbot];
	}
	// synchronisation point
	syncThreads();
	// refresh frontiers
	// two possible methods here:
	// 1. during the copy, every time a greater potential is found, do the reprop => adv: do reprop only on updated squares, disadv: a square is visited a lot of times (in the same way as depth-first search)
	// 2. first copy all the frontiers, and after do reprop on all squares
	// the 2nd method is used here
	ptsd->frontChanged[type] = 0;
	update = 0;
	if (ptsd->nfrontier)  // refresh N frontier
	  for (i=0 ; i<xbot-xtop ; i++)
	    if (pmgaGetSquareType(ytop, xtop+i) != WALLSQ){
	      max = ptsd->nfrontier[i];
	      if ((_nbDirs == 8) && (i != 0) && (ptsd->nfrontier[i-1] > max))
		max = ptsd->nfrontier[i-1];
	      if ((_nbDirs == 8) && (i != xbot-xtop-1) && (ptsd->nfrontier[i+1] > max))
		max = ptsd->nfrontier[i+1];
	      if (pots[ytop][xtop+i] + 1 < max)  // update square
		if ((pots[ytop][xtop+i] = max - 1) > 1)
		  update = 1;  // consider updated only if pot > 1
	    }
	if (ptsd->sfrontier)  // refresh S frontier
	  for (i=0 ; i<xbot-xtop ; i++)
	    if (pmgaGetSquareType(ybot-1, xtop+i) != WALLSQ){
	      max = ptsd->sfrontier[i];
	      if ((_nbDirs == 8) && (i != 0) && (ptsd->sfrontier[i-1] > max))
		max = ptsd->sfrontier[i-1];
	      if ((_nbDirs == 8) && (i != xbot-xtop-1) && (ptsd->sfrontier[i+1] > max))
		max = ptsd->sfrontier[i+1];
	      if (pots[ybot-1][xtop+i] + 1 < max)  // update square
		if ((pots[ybot-1][xtop+i] = max - 1) > 1)
		  update = 1;  // consider updated only if pot > 1
	    }
	if (ptsd->wfrontier)  // refresh W frontier
	  for (i=0 ; i<ybot-ytop ; i++)
	    if (pmgaGetSquareType(ytop+i, xtop) != WALLSQ){
	      max = ptsd->wfrontier[i];
	      if ((_nbDirs == 8) && (i != 0) && (ptsd->wfrontier[i-1] > max))
		max = ptsd->wfrontier[i-1];
	      if ((_nbDirs == 8) && (i != ybot-ytop-1) && (ptsd->wfrontier[i+1] > max))
		max = ptsd->wfrontier[i+1];
	      if (pots[ytop+i][xtop] + 1 < max)  // update square
		if ((pots[ytop+i][xtop] = max - 1) > 1)
		  update = 1;  // consider updated only if pot > 1
	    }
	if (ptsd->efrontier)  // refresh E frontier
	  for (i=0 ; i<ybot-ytop ; i++)
	    if (pmgaGetSquareType(ytop+i, xbot-1) != WALLSQ){
	      max = ptsd->efrontier[i];
	      if ((_nbDirs == 8) && (i != 0) && (ptsd->efrontier[i-1] > max))
		max = ptsd->efrontier[i-1];
	      if ((_nbDirs == 8) && (i != ybot-ytop-1) && (ptsd->efrontier[i+1] > max))
		max = ptsd->efrontier[i+1];
	      if (pots[ytop+i][xbot-1] + 1 < max)  // update square
		if ((pots[ytop+i][xbot-1] = max - 1) > 1)
		  update = 1;  // consider updated only if pot > 1
	    }
	if (update && (_nbDirs == 8))
	  ptsd->frontChanged[type] = 1;  // because it is possible to have /\ on the frontier
	// if necessary, refresh the environment with frontiers as sources
	VL (("Thread %d, frontierupdate=%d\n", ptsd->id, update));
	if (update)
	  repropfunc (type);
      }
  }while (loop);
}


/* *********************  1st method (recursive)  ******************* */


// propagate the potential of type TYPE starting in (y,x) with POTVALUE
// if the square already contains a greater value, it returns immediately.
// this recursive algorithm is depth-first search, which is not
// performant at all because a square is updated a lot of times for a
// given source.  Better use a breadth-first search algorithm
static void waveprop1 (int y, int x, int type, int potvalue)
{
  int dir;
  int **pot = gettype(type)->pots;
  tsd_t *ptsd;
  ICHECKCOORDS (y, x);
  if (pot[y][x] >= potvalue)
    return;  // there is already an equal or stronger potential
  pot[y][x] = potvalue;
  ptsd = getTSD();
  if (isOnDomainFrontier(y, x, ptsd->ytop, ptsd->xtop, ptsd->ybot, ptsd->xbot))
    ptsd->frontChanged[type] = 1;  // specify that the frontier was updated
  if (potvalue > 1)
    for (dir=0 ; dir<_nbDirs ; dir++){
      int newy = y + _diry[dir];
      int newx = x + _dirx[dir];
      if (isInMyDomain (newy, newx))
	if (pmgaGetSquareType(newy, newx) != WALLSQ)  // enters every square !WALL
	  waveprop1 (newy, newx, type, potvalue-1);
    }
}


// repropagate the frontiers
// wrapper for the prop function
static void wavereprop1 (int type)
{
  tsd_t *ptsd = getTSD();
  int ytop = ptsd->ytop, ybot = ptsd->ybot, xtop = ptsd->xtop, xbot = ptsd->xbot;
  int y, x, pot;
  int **pots = gettype(type)->pots;
  // reprop the N and S frontier
  for (x=xtop ; x<xbot ; x++){
    pot = pots[ytop][x];
    if (pot > 1){
      pots[ytop][x] = 0;
      waveprop1 (ytop, x, type, pot);
    }
    pot = pots[ybot-1][x];
    if (pot > 1){
      pots[ybot-1][x] = 0;
      waveprop1 (ybot-1, x, type, pot);
    }
  }
  // reprop the W and E frontier
  for (y=ytop+1 ; y<ybot-1 ; y++){  // +-1: these two squares have already been reprop
    pot = pots[y][xtop];
    if (pot > 1){
      pots[y][xtop] = 0;
      waveprop1 (y, xtop, type, pot);
    }
    pot = pots[y][xbot-1];
    if (pot > 1){
      pots[y][xbot-1] = 0;
      waveprop1 (y, xbot-1, type, pot);
    }
  }
}


// every thread visits every square in its domain to find all its mines,
// and propagates their field into its domain.
static void updateGradients1 (int updtype, void (*repropfunc)(int))
{
  int type, y, x;
  tsd_t *ptsd = getTSD();
  I ((updtype == USEPOT) || (updtype == USELOAD));
  // clear potentials
  clearPotentials ();
  // propagate the potential
  for (type=0 ; type<depthStack(&gss) ; type++)
    ptsd->frontChanged[type] = 0;  // clear the changed frontier flag
  for (y=ptsd->ytop ; y<ptsd->ybot ; y++)
    for (x=ptsd->xtop ; x<ptsd->xbot ; x++)
      if (pmgaGetSquareType(y, x) == RESSQ){  // propagate its potential
	srctype_t *ags;
	gradsrc_t *src;
	type = user2type(pmgaGetSquareResType(y, x));
	ags = gettype(type);
	src = getsrc(ags, getSquareSourceId (y, x));
	I (y == src->y);
	I (x == src->x);
	if (updtype == USELOAD)  // update the pot field of the gradient srcs
          if (src->func)
            src->pot = src->func(src->load, src->initload, src->initpot);
	//trop VL (("Thread %d: recGrad from %d, %d\n", me(), y, x));
	waveprop1 (y, x, type, src->pot);
      }
  // boundary exchange + repropagation (if needed)
  boundaryExchg (repropfunc);
  //VL (("Thread %d: potential updated, cycle %d updtype %d\n", ptsd->id, ptsd->cycle, updtype));
}


/* *****************  2nd method (iterative)  ****************** */


// TODO: replace MAX-es by smth optimized (because pots[...][...] as parameter of MAX are computed every time)
#define MAX2(x, y) (((x) > (y)) ? (x) : (y))
// TODO: implement more efficient MAX3 and MAX4
#define MAX3(x, y, z) (((x) > MAX2((y), (z))) ? (x) : MAX2((y), (z)))
#define MAX4(x, y, z, t) MAX2 (MAX2(x, y), MAX2(z, t))
// propagate the potential of this type

// I thought to write a common macro for the functions below, but there are pbs:
// - normy or not
// - normx or not
// - modified & setFrCh or not
// - MAX2, MAX3 or MAX4 (difficult) => I do not know how to write the fct.below

// update the potential of a square based on potential of its neighbour squares
#define UPDATE_SQUARE(neighb4, neighb8, frontier)			 \
do{									 \
  if (pmgaGetSquareType(y, x) != WALLSQ){  /* enter each square !WALL */ \
    /* safe, because the potential of the walls is zero: */		 \
    int pmax = (neighb4);						 \
    if (_nbDirs == 8){  /* 8-square propagation */			 \
      int ptmp = (neighb8);						 \
      if (pmax < ptmp)							 \
        pmax = ptmp;							 \
    }									 \
    if (pots[y][x] < pmax-1)						 \
      if ((pots[y][x] = pmax-1) > 1){					 \
        modified = 1;							 \
        frontier;							 \
      }									 \
  }									 \
}while (0)


// propagate entirely its domain
static void waveprop2 (int type)
{
  tsd_t *ptsd = getTSD();
  int y, x;
  int ytop = ptsd->ytop, ybot = ptsd->ybot, xtop = ptsd->xtop, xbot = ptsd->xbot;
  int **pots = gettype(type)->pots;
  int modified, frontchanged;
  X (int number = 0);
  frontchanged = 0;  // clear the changed frontier flag
  while (1){  // loop until the partition is entirely updated
    // from left to right and top to bottom
    X (number ++);
    modified = 0;
    y = ytop;
    // NW corner
    x = xtop;
    UPDATE_SQUARE (MAX2 (pots[y+1][x], pots[y][x+1]),
		   pots[y+1][x+1],
		   frontchanged = 1);
    // N frontier
    y = ytop;
    for (x=xtop+1 ; x<xbot-1 ; x++)
      UPDATE_SQUARE (MAX3 (pots[y+1][x], pots[y][x-1], pots[y][x+1]),
		     MAX2 (pots[y+1][x-1], pots[y+1][x+1]),
		     frontchanged = 1);
    // NE corner
    x = xbot - 1;
    UPDATE_SQUARE (MAX2 (pots[y+1][x], pots[y][x-1]),
		   pots[y+1][x-1],
		   frontchanged = 1);
    // internal lines
    for (y=ytop+1 ; y<ybot-1 ; y++){
      // W square
      x = xtop;
      UPDATE_SQUARE (MAX3 (pots[y-1][x], pots[y+1][x], pots[y][x+1]),
		     MAX2 (pots[y-1][x+1], pots[y+1][x+1]),
		     frontchanged = 1);
      // internal line
      for (x=xtop+1 ; x<xbot-1 ; x++)
	UPDATE_SQUARE (MAX4 (pots[y-1][x], pots[y+1][x], pots[y][x-1], pots[y][x+1]),
		       MAX4 (pots[y-1][x-1], pots[y-1][x+1], pots[y+1][x-1], pots[y+1][x+1]),
		       ;);  // here the frontier does not change
      // E square
      x = xbot - 1;
      UPDATE_SQUARE (MAX3 (pots[y-1][x], pots[y+1][x], pots[y][x-1]),
		     MAX2 (pots[y-1][x-1], pots[y+1][x-1]),
		     frontchanged = 1);
    }
    y = ybot - 1;
    // SW corner
    x = xtop;
    UPDATE_SQUARE (MAX2 (pots[y-1][x], pots[y][x+1]),
		   pots[y-1][x+1],
		   frontchanged = 1);
    // S frontier
    for (x=xtop+1 ; x<xbot-1 ; x++)
      UPDATE_SQUARE (MAX3 (pots[y-1][x], pots[y][x-1], pots[y][x+1]),
		     MAX2 (pots[y-1][x-1], pots[y-1][x+1]),
		     frontchanged = 1);
    // SE corner
    x = xbot - 1;
    UPDATE_SQUARE (MAX2 (pots[y-1][x], pots[y][x-1]),
		   pots[y-1][x-1],
		   frontchanged = 1);
    if (!modified)
      break;
    modified = 0;
    // from right to left and bottom to top
    X (number ++);
    y = ybot - 1;
    // SE corner
    x = xbot - 1;
    UPDATE_SQUARE (MAX2 (pots[y-1][x], pots[y][x-1]),
		   pots[y-1][x-1],
		   frontchanged = 1);
    // S frontier
    for (x=xbot-2 ; x>xtop ; x--)
      UPDATE_SQUARE (MAX3 (pots[y-1][x], pots[y][x-1], pots[y][x+1]),
		     MAX2 (pots[y-1][x-1], pots[y-1][x+1]),
		     frontchanged = 1);
    // SW corner
    x = xtop;
    UPDATE_SQUARE (MAX2 (pots[y-1][x], pots[y][x+1]),
		   pots[y-1][x+1],
		   frontchanged = 1);
    // internal lines
    for (y=ybot-2 ; y>ytop ; y--){
      // E square
      x = xbot - 1;
      UPDATE_SQUARE (MAX3 (pots[y-1][x], pots[y+1][x], pots[y][x-1]),
		     MAX2 (pots[y-1][x-1], pots[y+1][x-1]),
		     frontchanged = 1);
      // internal line
      for (x=xbot-2 ; x>xtop ; x--)
	UPDATE_SQUARE (MAX4 (pots[y-1][x], pots[y+1][x], pots[y][x-1], pots[y][x+1]),
		       MAX4 (pots[y-1][x-1], pots[y-1][x+1], pots[y+1][x-1], pots[y+1][x+1]),
		       ;);  // here the frontier does not change
      // W square
      x = xtop;
      UPDATE_SQUARE (MAX3 (pots[y-1][x], pots[y+1][x], pots[y][x+1]),
		     MAX2 (pots[y-1][x+1], pots[y+1][x+1]),
		     frontchanged = 1);
    }
    y = ytop;
    // NE corner
    x = xbot - 1;
    UPDATE_SQUARE (MAX2 (pots[y+1][x], pots[y][x-1]),
		   pots[y+1][x-1],
		   frontchanged = 1);
    // N frontier
    for (x=xbot-2 ; x>xtop ; x--)
      UPDATE_SQUARE (MAX3 (pots[y+1][x], pots[y][x-1], pots[y][x+1]),
		     MAX2 (pots[y+1][x-1], pots[y+1][x+1]),
		     frontchanged = 1);
    // NW corner
    x = xtop;
    UPDATE_SQUARE (MAX2 (pots[y+1][x], pots[y][x+1]),
		   pots[y+1][x+1],
		   frontchanged = 1);
    if (!modified)
      break;
  }
  ptsd->frontChanged[type] = frontchanged;
  VL (("Thread %d, number of sweeps=%d\n", ptsd->id, number));
}


// iterative method to update gradients (4 or 8 directions)
// (for every square, its pot = max(its pot, the max value of its neigh - 1))
// this method is hoping to be more efficient in execution time than
// updateGradients() in case there are more than a few resources
static void updateGradients2 (int updtype, void (*repropfunc)(int))
{
  int type, y, x;
  tsd_t *ptsd = getTSD();
  I ((updtype == USEPOT) || (updtype == USELOAD));
  // clear potentials
  clearPotentials ();
  // set potential of resource squares with the resource potential
  for (y=ptsd->ytop ; y<ptsd->ybot ; y++)
    for (x=ptsd->xtop ; x<ptsd->xbot ; x++)
      if (pmgaGetSquareType(y, x) == RESSQ){
	srctype_t *ags;
	gradsrc_t *src;
	type = user2type(pmgaGetSquareResType (y, x));
	ags = gettype(type);
	src = getsrc(ags, getSquareSourceId (y, x));
	I (y == src->y);
	I (x == src->x);
	if (updtype == USELOAD)  // update the pot field of the gradient srcs
          if (src->func)
            src->pot = src->func(src->load, src->initload, src->initpot);
	ags->pots[y][x] = src->pot;
      }
  // propagate the potential
  for (type=0 ; type<depthStack(&gss) ; type++)
    waveprop2 (type);
  // boundary exchange + repropagation (if needed)
  boundaryExchg (repropfunc);
  //VL (("Thread %d: potential updated, cycle %d updtype %d\n", ptsd->id, ptsd->cycle, updtype));
}


/*****  3rd method (manually recursive and more optimised than the 1st)  ****/


// propagate the potential of type TYPE starting in (y,x) with CRTPOT
// Pre-condition: outside exit condition (crtpot > pot[y][x])
// this manually recursive algorithm is breadth-first search and uses a queue
// (compare with the 1st sol)
// unfortunately, during the propagation, it seems impossible to avoid
//   to re-check the square where it comes from
// Credits to Herve (who had the idea and explanation of breadth-first search)
static void waveprop3 (int y, int x, int type, int crtpot)
{
  int dir;
  int **pot = gettype(type)->pots;
  tsd_t *ptsd = getTSD();
  int ytop = ptsd->ytop, ybot = ptsd->ybot, xtop = ptsd->xtop, xbot = ptsd->xbot;
  queue_t *psrc;
  ICHECKCOORDS (y, x);
  I (crtpot >= 1);
  I (crtpot > pot[y][x]);  // outside exit condition
  psrc = &ptsd->srcQueue;
  // max field (2 ints/square) needed and
  // reset the index to 0 (the queue is not circular)
  resetQueue (psrc, 2 * maxfield (crtpot));
  pot[y][x] = crtpot;  // before every addints we mark the point (this solved a bug!)
  addints (psrc, y, x);
  crtpot--;  // we start marking with crtpot-1;
  addints (psrc, DELIM, DELIM);
  while (1){
    removeints (psrc, &y, &x);
    if (y == DELIM){
      crtpot--;
      if ((crtpot == 0) || (depthQueue(psrc) == 0))  // end of prop
	break;
      addints (psrc, DELIM, DELIM);
      removeints (psrc, &y, &x);
      I (y != DELIM);
    }
    if (isOnDomainFrontier(y, x, ytop, xtop, ybot, xbot)){  // use isInMyDomain
      ptsd->frontChanged[type] = 1;  // the frontier was updated.  NB: only if its pot is > 1
      for (dir=0 ; dir<_nbDirs ; dir++){
	register int newy = y + _diry[dir];
	register int newx = x + _dirx[dir];
	// outside exit conditions (optimisation)
	if (isInMyDomain_ (newy, newx, ytop, xtop, ybot, xbot)
	    && (pmgaGetSquareType(newy, newx) != WALLSQ)
	    && (pot[newy][newx] < crtpot)){
	  pot[newy][newx] = crtpot;
	  // the `if' grows exponentially,
	  // the squares to add to the queue only linear
	  // But an `if' is many times slower than an addition to the queue
	  if (crtpot > 1)  // we can better move this line outside the for
	    addints (psrc, newy, newx);
	}
      }
    }else  // NB: not using isInMyDomain gives a very high improvement on texec
      for (dir=0 ; dir<_nbDirs ; dir++){
	register int newy = y + _diry[dir];
	register int newx = x + _dirx[dir];
	// outside exit condition (optimisation)
	if ((pmgaGetSquareType(newy, newx) != WALLSQ)
	    && (pot[newy][newx] < crtpot)){
	  pot[newy][newx] = crtpot;
	  // the `if' grows exponentially,
	  // the squares to add to the queue only linear
	  // But an `if' is many times slower than an addition to the queue
	  if (crtpot > 1)  // we can better move this line outside the for
	    addints (psrc, newy, newx);
	}
      }
  }
}


// Optimisation: if the next square on the frontier contains a higher
// value, then do not propagate current square (since it will be
// propagated anyway), but set it to 0 and go to the next square.
// repropagate the N and S frontiers
#define REPROPY(top)                                            \
for (x=xtop ; x<xbot ; x++){                                    \
  pot = pots[(top)][x];                                         \
  /* next `if' is an optimisation */                            \
  if ((x < xbot-1) && (pot < pots[(top)][x+1])){                \
    pots[(top)][x] = 0;  /* it will be regenerated anyway */    \
    continue;                                                   \
  }                                                             \
  if (pot > 1){                                                 \
    pots[(top)][x] = 0;                                         \
    waveprop3 ((top), x, type, pot);                            \
  }                                                             \
}

// repropagate the E and W frontiers
#define REPROPX(top)                                            \
/* +-1: these two squares have already been reprop */           \
for (y=ytop+1 ; y<ybot-1 ; y++){                                \
  pot = pots[y][(top)];                                         \
  /* next `if' is an optimisation */                            \
  if ((y < ybot-2) && (pot < pots[y+1][(top)])){                \
    pots[y][(top)] = 0;  /* it will be regenerated anyway */    \
    continue;                                                   \
  }                                                             \
  if (pot > 1){                                                 \
    pots[y][(top)] = 0;                                         \
    waveprop3 (y, (top), type, pot);                            \
  }                                                             \
}

// repropagate the frontiers
// wrapper for the prop function
static void wavereprop3 (int type)
{
  tsd_t *ptsd = getTSD();
  int ytop = ptsd->ytop, ybot = ptsd->ybot, xtop = ptsd->xtop, xbot = ptsd->xbot;
  int y, x, pot;
  int **pots = gettype(type)->pots;
  REPROPY (ytop);  // reprop N frontier
  REPROPY (ybot-1);  // reprop S frontier
  REPROPX (xtop);  // reprop W frontier
  REPROPX (xbot-1);  // reprop E frontier
}


static void updateGradients3 (int updtype, void (*repropfunc)(int))
{
  int type, y, x;
  tsd_t *ptsd = getTSD();
  int ytop = ptsd->ytop, ybot = ptsd->ybot, xtop = ptsd->xtop, xbot = ptsd->xbot;
  I ((updtype == USEPOT) || (updtype == USELOAD));
  // clear potentials
  clearPotentials ();
  // set potential of resource squares with the resource pot and propagate
  for (type=0 ; type<depthStack(&gss) ; type++){
    srctype_t *ags = gettype(type);
    ptsd->frontChanged[type] = 0;
    for (y=ytop ; y<ybot ; y++)
      for (x=xtop ; x<xbot ; x++)
	if ((pmgaGetSquareType(y,x) == RESSQ) && (pmgaGetSquareResType(y,x) == type2user(type))){
	  gradsrc_t *src = getsrc(ags, getSquareSourceId (y, x));
	  I (y == src->y);
	  I (x == src->x);
	  if (updtype == USELOAD)  // update the pot field of the grad srcs
            if (src->func)
              src->pot = src->func(src->load, src->initload, src->initpot);
	  if (ags->pots[y][x] < src->pot)  // outside exit cond
	    waveprop3 (y, x, type, src->pot);
	}
  }
  // boundary exchange + repropagation (if needed)
  boundaryExchg (repropfunc);
  //VL (("Thread %d: potential updated, cycle %d updtype %d\n", ptsd->id, ptsd->cycle, updtype));
}


/* ****************  self method: private domains  ****************** */
// This will be changed with a method which can use any of the seq methods

static void privateDomains (int updtype)
{
  tsd_t *ptsd = getTSD();
  int type, y, x;
  int modified;
  int **pots = globalPots[ptsd->id];  // used for all the types
  I ((updtype == USEPOT) || (updtype == USELOAD));
  if (pots == NULL){
    pots = malloc (_dimy * sizeof(int*));
    if (pots == NULL)
      abcerror ("Cannot allocate %d bytes in privateDomains.  Aborting.\n", _dimy * sizeof(int*));
    pots[0] = malloc (_dimy * _dimx * sizeof(int));
    if (pots[0] == NULL)
      abcerror ("Cannot allocate %d bytes in privateDomains.  Aborting.\n", _dimy * _dimx * sizeof(int));
    for (y=1 ; y<_dimy ; y++)
      pots[y] = pots[0] + y * _dimx;
    globalPots[ptsd->id] = pots;
  }
  for (type=0 ; type<depthStack(&gss) ; type++){
    srctype_t *ags = gettype(type);
    // clear potentials
    memset (pots[0], 0, _dimy * _dimx * sizeof(int));
    // set potential of resource squares
    for (y=ptsd->ytop ; y<ptsd->ybot ; y++)
      for (x=ptsd->xtop ; x<ptsd->xbot ; x++)
	if ((pmgaGetSquareType(y, x) == RESSQ) && (pmgaGetSquareResType(y,x) == type2user(type))){
	  gradsrc_t *src = getsrc(ags, getSquareSourceId (y, x));
	  I (y == src->y);
	  I (x == src->x);
	  if (updtype == USELOAD)  // update the pot field of the gradient srcs
            if (src->func)
              src->pot = src->func(src->load, src->initload, src->initpot);
	  pots[y][x] = src->pot;
	}
    // propagate into the private domain
    while (1){  // loop until the partition is entirely updated
      modified = 0;
      for (y=0 ; y<_dimy ; y++)
	for (x=0 ; x<_dimx ; x++)
	  if (pmgaGetSquareType(y, x) != WALLSQ){  // enters every square !WALL
	    // safe, because the potential of the walls is zero:
	    int pmax = MAX4 (pots[y-1][x], pots[y+1][x],
			     pots[y][x-1], pots[y][x+1]);
	    if (_nbDirs == 8){  // 8-square propagation
	      int ptmp = MAX4 (pots[y-1][x-1], pots[y-1][x+1],
			       pots[y+1][x-1], pots[y+1][x+1]);
	      if (pmax < ptmp)
		pmax = ptmp;
	    }
	    if (pots[y][x] < pmax-1){
	      pots[y][x] = pmax-1;
	      modified = 1;
	    }
	  }
      if (!modified)
	break;
      modified = 0;
      for (y=_dimy-1 ; y>=0 ; y--)
	for (x=_dimx-1 ; x>=0 ; x--)
	  if (pmgaGetSquareType(y, x) != WALLSQ){  // enters every square !WALL
	    // safe, because the potential of the walls is zero:
	    int pmax = MAX4 (pots[y-1][x], pots[y+1][x],
			     pots[y][x-1], pots[y][x+1]);
	    if (_nbDirs == 8){  // 8-square propagation
	      int ptmp = MAX4 (pots[y-1][x-1], pots[y-1][x+1],
			       pots[y+1][x-1], pots[y+1][x+1]);
	      if (pmax < ptmp)
		pmax = ptmp;
	    }
	    if (pots[y][x] < pmax-1){
	      pots[y][x] = pmax-1;
	      modified = 1;
	    }
	  }
      if (!modified)
	break;
    }
    // synchronize threads
    syncThreads ();
    // update the main matrix
    for (y=ptsd->ytop ; y<ptsd->ybot ; y++)
      for (x=ptsd->xtop ; x<ptsd->xbot ; x++){
	int i, max = 0;
	for (i=0 ; i<_nbThreads ; i++){
	  int thpot = globalPots[i][y][x];
	  if (max < thpot)
	    max = thpot;
	}
	ags->pots[y][x] = max;
      }
    syncThreads ();  // wait until all globalPots is read.  I could use a 2D matrix to avoid this sync, but this takes memory...
  }
  //VL (("Thread %d: potential updated, cycle %d updtype %d\n", ptsd->id, ptsd->cycle, updtype));
}


/* *******  private domains with breadth-first recursive propagation   ***** */
// This part is COPIED from privateDomains and recursive breadth-first

// Some parts can be removed for privateDomainsBFR method, but I prefered
// to leave them in order to have the same texec than original BFR method
static void waveprop3BFRecursive (int y, int x, int type, int crtpot)
{
  int dir;
  tsd_t *ptsd = getTSD();
  int **pot = globalPots[ptsd->id];
  int ytop = ptsd->ytop, ybot = ptsd->ybot, xtop = ptsd->xtop, xbot = ptsd->xbot;
  queue_t *psrc;
  ICHECKCOORDS (y, x);
  I (crtpot >= 1);
  I (crtpot > pot[y][x]);  // outside exit condition
  psrc = &ptsd->srcQueue;
  // max field (2 ints/square) needed and
  // reset the index to 0 (the queue is not circular)
  resetQueue (psrc, 2 * maxfield (crtpot));
  pot[y][x] = crtpot;  // before every addints we mark the point (this solved a bug!)
  addints (psrc, y, x);
  crtpot--;  // we start marking with crtpot-1;
  addints (psrc, DELIM, DELIM);
  while (1){
    removeints (psrc, &y, &x);
    if (y == DELIM){
      crtpot--;
      if ((crtpot == 0) || (depthQueue(psrc) == 0))  // end of prop
	break;
      addints (psrc, DELIM, DELIM);
      removeints (psrc, &y, &x);
      I (y != DELIM);
    }
    if (isOnDomainFrontier(y, x, ytop, xtop, ybot, xbot))
      ptsd->frontChanged[type] = 1;  // the frontier was updated.  NB: only if its pot is > 1
    for (dir=0 ; dir<_nbDirs ; dir++){
      register int newy = y + _diry[dir];
      register int newx = x + _dirx[dir];
      // check for isInDomain_ removed, because unnecessary
      // outside exit condition (optimisation)
      if ((pmgaGetSquareType(newy, newx) != WALLSQ)
	  && (pot[newy][newx] < crtpot)){
	pot[newy][newx] = crtpot;
	// the `if' grows exponentially,
	// the squares to add to the queue only linear
	// But an `if' is many times slower than an addition to the queue
	if (crtpot > 1)  // we can better move this line outside the for
	  addints (psrc, newy, newx);
      }
    }
  }
}

static void privateDomainsBFRecursive (int updtype)
{
  tsd_t *ptsd = getTSD();
  int type, y, x;
  int **pots = globalPots[ptsd->id];  // used for all the types
  I ((updtype == USEPOT) || (updtype == USELOAD));
  if (pots == NULL){
    pots = malloc (_dimy * sizeof(int*));
    if (pots == NULL)
      abcerror ("Cannot allocate %d bytes in privateDomainsBFRecursive.  Aborting.\n", _dimy * sizeof(int*));
    pots[0] = malloc (_dimy * _dimx * sizeof(int));
    if (pots[0] == NULL)
      abcerror ("Cannot allocate %d bytes in privateDomains.  Aborting.\n", _dimy * _dimx * sizeof(int));
    for (y=1 ; y<_dimy ; y++)
      pots[y] = pots[0] + y * _dimx;
    globalPots[ptsd->id] = pots;
  }
  for (type=0 ; type<depthStack(&gss) ; type++){
    srctype_t *ags = gettype(type);
    // clear potentials
    memset (pots[0], 0, _dimy * _dimx * sizeof(int));
    // propagate resources
    for (y=ptsd->ytop ; y<ptsd->ybot ; y++)
      for (x=ptsd->xtop ; x<ptsd->xbot ; x++)
	if ((pmgaGetSquareType(y, x) == RESSQ) && (pmgaGetSquareResType(y,x) == type2user(type))){
	  gradsrc_t *src = getsrc(ags, getSquareSourceId (y, x));
	  I (y == src->y);
	  I (x == src->x);
	  if (updtype == USELOAD)  // update the pot field of the gradient srcs
            if (src->func)
              src->pot = src->func(src->load, src->initload, src->initpot);
	  // two lines added for BF and one commented
	  if (pots[y][x] < src->pot)  // outside exit cond
	    waveprop3BFRecursive (y, x, type, src->pot);
	  //	  pots[y][x] = src->pot;
	}
    // synchronize threads
    syncThreads ();
    // update the main matrix
    for (y=ptsd->ytop ; y<ptsd->ybot ; y++)
      for (x=ptsd->xtop ; x<ptsd->xbot ; x++){
	int i, max = 0;
	for (i=0 ; i<_nbThreads ; i++){
	  int thpot = globalPots[i][y][x];
	  if (max < thpot)
	    max = thpot;
	}
	ags->pots[y][x] = max;
      }
    syncThreads ();  // wait until all globalPots is read.  I could use a 2D matrix to avoid this sync, but this takes memory...
  }
  //VL (("Thread %d: potential updated, cycle %d updtype %d\n", ptsd->id, ptsd->cycle, updtype));
}



/* ******************  propagation methods  *********************** */

// set the propagation method
void pmSetPropagationMethod (int method)
{
  I ((method == BREADTH_BREADTH_DOMAIN)
     || (method == BREADTH_DEPTH_DOMAIN)
     || (method == BREADTH_ITERATIVE_DOMAIN)
     || (method == DEPTH_BREADTH_DOMAIN)
     || (method == DEPTH_DEPTH_DOMAIN)
     || (method == DEPTH_ITERATIVE_DOMAIN)
     || (method == ITERATIVE_BREADTH_DOMAIN)
     || (method == ITERATIVE_DEPTH_DOMAIN)
     || (method == ITERATIVE_ITERATIVE_DOMAIN)
     || (method == BREADTH_PRIVATE)
     || (method == ITERATIVE_PRIVATE));
  propMethod = method;
}


// updateGradients* (int updtype)
// if updtype == USEPOT, then pot field is used directly
// if updtype == USELOAD, then, before use, the pot is computed from the load
// NB:syncThreads is guaranteed (correct grad prop) at the end of this function
void updateGradients (int updtype)
{
  if (_envType == TORUS){  // wave propagation not yet implemented
    static int printed;
    if (!printed){
      fprintf (stderr, "Propagation of potential not done, since it is not yet implemented in torus environment.\n");
      printed = 1;
    }
    return;
  }
  switch (propMethod){
  case BREADTH_BREADTH_DOMAIN:
    updateGradients3 (updtype, wavereprop3);
    break;
  case BREADTH_DEPTH_DOMAIN:
    updateGradients3 (updtype, wavereprop1);
    break;
  case BREADTH_ITERATIVE_DOMAIN:
    updateGradients3 (updtype, waveprop2);
    break;
  case DEPTH_BREADTH_DOMAIN:
    updateGradients1 (updtype, wavereprop3);
    break;
  case DEPTH_DEPTH_DOMAIN:
    updateGradients1 (updtype, wavereprop1);
    break;
  case DEPTH_ITERATIVE_DOMAIN:
    updateGradients1 (updtype, waveprop2);
    break;
  case ITERATIVE_BREADTH_DOMAIN:
    updateGradients2 (updtype, wavereprop3);
    break;
  case ITERATIVE_DEPTH_DOMAIN:
    updateGradients2 (updtype, wavereprop1);
    break;
  case ITERATIVE_ITERATIVE_DOMAIN:
    updateGradients2 (updtype, waveprop2);
    break;
  case BREADTH_PRIVATE:
    privateDomainsBFRecursive (updtype);
    break;
  case ITERATIVE_PRIVATE:
    privateDomains (updtype);
    break;
  default:
    abort();  // library error
  }
}
