// declarations internal to this library

#include <time.h>
#include <sys/times.h>
#include "nana.h"

// constants

// used by updateGradients()
#define USEPOT  0  // use directly pot
#define USELOAD 1  // compute pot using load, then use pot

#define NOTRUNC 0  // both used for stat file
#define TRUNC   1  // truncate the file


// type definitions

typedef struct{
  int nbAllocElem;// number of allocated elements (>= nbElem)
  int nbElem;     // number of elements currently stocked
  int increment;  // number of elements to allocate when need to grow the stack
  void *data;     // pointer to the beginning of the data
}stack_t;

typedef struct{
  int lastIndex;  // lastIndex - queue growing with resetQueue
  int addIndex, removeIndex;  // the indexes of the first and the last element
  int *data;      // pointer to the beginning of the data
}queue_t;

// agentinfo structure (used in pmgaCreateAgent)
typedef struct{
  int y, x;
  int slowness;
  int priority;
  int maxload;
  void *memory;
  pmpfAgent func;
  int parentid;  // the id of the parent, used for reproducibility
  int created;  // check that every globally created agent is really dealt
}agentinfo_t;


// global variables

extern int _diry[];
extern int _dirx[];
extern int _dimy, _dimx;       // env dimensions, from 0 to _dimx-1
extern int _nbDirs;            // number of directions (4 or 8)
extern int _nbThreads;         // *total* number of threads in the application
extern unsigned int _gSeed;    // global seed, used for (seq) initialization
extern int _envType;           // the type of the environment (TORUS or EDGES)
extern stack_t _createdAgents;


// macros

// macro which checks that (y,x) is inside the environment
// the do-while loop is used to ensure that it is equivalent to a single stmt
#define ICHECKCOORDS(y, x)						     \
        do{								     \
          I (((y) >= 0) && ((y) < _dimy));  I (((x) >= 0) && ((x) < _dimx)); \
        }while (0)

// macro which checks that (y,x) is inside the environment
// the do-while loop is used to ensure that it is equivalent to a single stmt
#define ICHECKDIR(dir)					\
        do{						\
          I (((dir) >= 0) && ((dir) < _nbDirs));	\
        }while (0)

// returns a number depending on global seed, space & time
#define UNIQUE(cycle, y, x) (unsigned int)(_gSeed + 7*cycle + (y) * _dimx + x)

#define BENCH(funccall, threadno)					\
{									\
  clock_t c1, c2, t1, t2;						\
  struct tms ts1, ts2;							\
  t1 = times (&ts1);							\
  c1 = clock();								\
  funccall;								\
  c2 = clock();								\
  t2 = times (&ts2);							\
  printf (#funccall "clock: thread %d, diff=%lldms\n", threadno,	\
	  ((long long)c2-c1) * 1000 / CLOCKS_PER_SEC);			\
  printf (#funccall "times: thread %d, diff=%lldms\n", threadno,	\
	  ((long long)t2-t1) * 1000 / CLK_TCK);				\
}


// inline function definitions

// the both are taken from env.c
// normalize y, ie. returns y between 0 and _dimy-1 in TORUS case,
// else fails if y not in interval
// static => hope it is inlined
// The 2nd method works only for -dim..2dim-1
static int normy (int y)
{
/*   if (_envType == EDGES){ */
/*     I ((y >= 0) && (y < _dimy)); */
/*     return y;  // better perf (for gcc at least) - bug gcc? */
/*   } */
  // y = (y + _dimy) % _dimy;  // 1st method.  Works for -dim..oo
  I ((y >= -_dimy) && (y < 2*_dimy));
  if (y < 0)  // 2nd method.  Works for -dim..2dim-1, but is the fastest
    y += _dimy;
  else if (y >= _dimy)
    y -= _dimy;
/*   y = y % _dimy;  // 3rd method (general) */
/*   // if y was < 0, % returns a negative value, whereas y must be >0 */
/*   if (y < 0) */
/*     y += _dimy; */
  return y;
}

// normalize x, ie. returns x between 0 and _dimx-1 in TORUS case,
// else fails if x not in interval
// static => hope it is inlined
// The 2nd method works only for -dim..2dim-1
static int normx (int x)
{
/*   if (_envType == EDGES){ */
/*     I ((x >= 0) && (x < _dimx)); */
/*     return x;  // better perf (for gcc at least) - bug gcc? */
/*   } */
  // x = (x + _dimx) % _dimx;
  I ((x >= -_dimx) && (x < 2*_dimx));
  if (x < 0)
    x += _dimx;
  else if (x >= _dimx)
    x -= _dimx;
/*   x = x % _dimx; */
/*   // if x was < 0, % returns a negative value, whereas x must be >0 */
/*   if (x < 0) */
/*     x += _dimx; */
  return x;
}


// function declarations

#ifndef WITHOUT_NANA
#define X(s) s
#else
#define X(s) /* empty */
#endif

// system.c
void setSeed       (unsigned int);     // random numbers are used by agents
void setInitialCycle (int);

// env.c
void setSquareState (int, int, int);   // y, x, state
void putSquareGS (int, int, int, int); // y, x, type, id: this func sets type, id of gradsrc, and sqtype
void setSquareAgentId (int, int, int);  // y, x, id
int getSquareSourceType (int, int);    // y, x
int getSquareSourceId (int, int);      // y, x
void freeEnv (void);
int getObstNumber (void);

// saveChanges.c
void closeChangesFiles (void);
void endStat (void);
int getChangesFileName (char *);
int initChangesAppend (char *);

// saveSys.c
void initSaveSys (void);
void freeSaveSys (void);
void setParallelSavingAllowed (int);
void setParallelSavingDelayed (int);
int getParallelSavingDelayed (void);
int saveLocalFilePar (int);

// saveStat.c
int getStatFileName (char *);
int initStatAppend (char *);  // filename for statistic file, open for appending

// gradsrc.c
void initResource (void);
void initResource2 (void);
void initResourcePar (void);
int type2user (int);
void updateGradients (int);  // type
void freeResource (void);
void freeResourcePar (void);
int needupdate (void);
int getResourcePeriod (void);
int getUnvisitedResNumber (int);  // type
int getEmptyResNumber (int);  // type
int getResTotalLoad (int);  // type
int getInitResLoad (void);    // type
void changeResourceLoad (int, int, int);  // type, id, changeload
int getResourceLoad (int, int);  // type, id
int getResourceInitLoad (int, int);  // type, id
int getResourcePot (int, int); // type, id
int getResourceInitPot (int, int);  // type, id
int getResourcey (int, int);   // type, id
int getResourcex (int, int);   // type, id
pmpfSource getResourceFunc (int, int);  // type, id
int getResourceNumber (int);   // type
int getResourceTypesNumber (void);
void putUsertypeResource (int, int, int, int, int, int, int, pmpfSource);
int getTotalResourceNumber (void);

// agent.c
void initAgent (void);
void initAgentPar (void);
void freeAgent (void);
int isAlive (int);
int isKilled (int);
int getRandomAgentNumber (void);
void clearRandomAgent (void);
int getAgentNumbermax (void);
int getEmptyAgentNumber (void);
int getAgentSlowness (int);  // id
int getAgentPriority (int);  // id
pmpfAgent getAgentFunc (int);  // id
void *getAgentMemory (int);  // id
void planAgentMove (int, int);  // id
void doAgentMove (int);  // id
void doAgentTake (int);  // id
void doAgentDrop (int);  // id
void setAgentActiontype (int, int);  // id, actiontype
void setAgentAction (int, pmAction_t);  // id, pmAction_t
int getAgentnewy (int);  // id
int getAgentnewx (int);  // id
void putAgentAll (int, int, int, int, int, int, void*, int, pmpfAgent);
int getAgentRandomMove (int, int *);  // agent, &dir
void processWinnerCA (int);
void processLooserCA (int);
void markDeadAgent (int);
//void clearMAgent (void);
//int getMAgentNumber (void);
//void addMAgent (int);  // agent
//int getMAgent (int);  // agent
// agent.c
#ifndef WITHOUT_NANA
void testAndsetAgentplan (int);
void testAndsetAgentmove (int);
void testAndclearAgents (void);
void clearAgents (void);
void testAndsetMAgent (int);
void testAndclearMAgent (int);
void testAndclearCAgents (stack_t *);
#endif

// conflicts.c
void initConflictTablePar (void);
void clearConflictTablePar (void);
void freeConflictTablePar (void);
int addConflictStructurePar (void);
int getConflictsNumberPar (void);
void addConflictItemPar (int, int);  // cfNumber, item
void addConflictItemCAPar (int, int);
int getConflictItemsNumberPar (int);  // cfNumber
int getConflictItemsNumberCAPar (int);
int getConflictDataPar (int, int);  // cfNumber, item
int getConflictDataCAPar (int, int);
void sortConflictStructurePar (int);  // cfNumber
void sortConflictStructureCAPar (int);

// newpos.c
void initNewPos (int, int);  // y, x
void clearNewPos (void);
void addNewPos (int, int, int);  // y, x, objid
void addNewPosCA (int, int, int);  // y, x, objid
void freeNewPos (void);

// misc.c
int globrand (int);  // 0..nb-1
void abcerror (char *, int);  // prints a memory error and aborts
int isInMyDomain (int, int);  // y, x

// debug.c
void printAgentInfo (void);
void printConflictsInfo (void);

// vision.c
void allocVisionMatrix (void);
void allocVisionMatrixPar (void);
void fillVisionMatrixPar (void);
void freeVisionMatrix (void);
void freeVisionMatrixPar (void);
int getVisionRadius (void);
int getVisionType (void);

// stack.c
void initStack (stack_t *, int);
void pushint (stack_t *, int);
int popint (stack_t *);
int peekint (stack_t *, int);  // index
int depthStack (stack_t *);
void clearStack (stack_t *);
void endStack (stack_t *);
int  pushstruct (stack_t *, void*, int);  // elem, size.  Return the new index
void popstruct (stack_t *, void*, int);  // elem, size
void peekstruct (stack_t *, int, void*, int);  // index, elem, size
void *peekstructp (stack_t *, int, int);  // index, size

// queue.c
void initQueue (queue_t *, int);
void addints (queue_t *, int, int);
void removeints (queue_t *, int *, int *);
int depthQueue (queue_t *);
void resetQueue (queue_t *, int);  // needed size
void endQueue (queue_t *);
