// This file deals with save/load of the whole system.
// This means the system may be resumed after being restored
//   (in the exactly the same configuration).
// To rewrite the load part with lex/yacc.
// Take care: the system must be coherent (not updated partially) when saving,
//   because loading does an updatePotential().
// To take care, not re-entrant functions I think.

// if stdio.h does not define FILENAME_MAX, then we have to define it ourselves
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>  // only for time save in saveGeneralFile (may be removed if not portable)
#include "agents.h"
#include "intern.h"
#include "lowinterface.h"
#include "nana.h"

/*
  A file with general information (global), and a file per thread (local).
  Two solutions of local file:
  1. either save walls first, agents after, and src last (without abbr)
  2. or save in environment order (left->right) or random (with abbr)
    abbr is O, A, or S, and is used to know what that line is about
  The second solution is implemented.

  Requirements:
  1. Comments: # or cpp use (first style implemented)
  2. keyword-key (changeable), or key (inchangeable) (first style implemented)
  3. lisible files, which may be edited manually
  4. space or performance less important

  1 global file
  f local files (assume every file uses its domain)
  p threads

  Actual limitations:
  - max 1024 chars/line (with checking, but cuts the string)

  Parse issues (G-global file, L-local file):
  - GL any order of lines
  - GL # starts comments
  - GL missing numbers are checked (only in local)
  - G  missing or double fields are checked with error return value
*/

#define MAXEXT      5
#define MAXLGPROC   4        // suppose max number of threads <= 10^4

// abbreviations in the files
#define OBST_ABBR "W"
#define AGENT_ABBR "A"
#define RES_ABBR "S"

// formatted strings for the local files
#define OBST_STRING "%d\t%d"                         // y x
#define AGENT_STRING "%d\t%d\t%d\t%d\t%d\t%d\t%d"    // y x slowness priority load maxload func_id
#define RES_STRING "%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d"  // y x type load initload pot initpot func_id

// formatted string for the general file
#define GENFILE_STRING "#start with cycle number...\ncycle %d\n\n#number of threads used\nnbThreads %d\n\n#global seed (an unsigned int)\ngseed %u\n\n#number of lines of the environment\ndimy %d\n\n#number of columns of the environment\ndimx %d\n\n#type of the environment: 0=TORUS, 1=EDGES\nenvtype %d\n\n#number of directions used: 4 or 8\nnbdirs %d\n\n#period of gradient sources updates: 0=initialization only, k=every k cycles\ngradsrcUpdate %d\n\n#vision radius: 0=no vision, n=(radius=n)\nvisionradius %d\n\n#vision type (0=STATIC, 1=DYNAMIC)\nvisiontype %d\n\n#statistics file name\nstatfilename %s\n\n#statistics already saved? (0=no, 1=yes)\nstatsaved %d\n\n#changes file name\nchangesfilename %s\n\n#changes already saved? (0=no, 1=yes)\nchangessaved %d\n\n#local files (ytop xtop filename) - coordinates start with 0\n"

#define INCFILES 5          // increment by 5 the stack of local files
#define MAXFUNCS 100        // max number of functions which can be stored

typedef struct{
  int ytop;
  int xtop;
  char name[FILENAME_MAX];
}localfile_t;
static stack_t localfiles;

static int bParallelSaving; // will the saving of local files be parallelized?
static int bParallelSavingAllowed;  // can the saving be now done in parallel?
static int bParallelSavingDelayed;  // have to save the local files

static pmpfAgent agfuncsSave[MAXFUNCS];  // used by pmSaveSystem
static pmpfSource l2pfuncsSave[MAXFUNCS];  // TODO: use a dynamic allocation
static char filenameSave[FILENAME_MAX];
// NB: if you want to save for load too, you must use another set of variables, to not conflict with these three

void initSaveSys (void)
{
  initStack (&localfiles, INCFILES);
  bParallelSaving = 1;      // parallel saving by default
  bParallelSavingAllowed = 0;  // parallel saving works only inside runfunc
  bParallelSavingDelayed = 0;  // no saving in execution
}

void freeSaveSys (void)
{
  endStack (&localfiles);
}

// or two funcs: set & reset?
void pmSetParallelSaving (int i)
{
  I ((i == 0) || (i == 1));
  bParallelSaving = i;
}

// NB: when calling this function with 1, call also setParallelSavingDelayed(0)
void setParallelSavingAllowed (int i)
{
  I ((i == 0) || (i == 1));
  bParallelSavingAllowed = i;
}

void setParallelSavingDelayed (int i)
{
  I ((i == 0) || (i == 1));
  bParallelSavingDelayed = i;
}

int getParallelSavingDelayed (void)
{
  return bParallelSavingDelayed;
}

// ---------------------------------------------

// put in lfn the local file name, based on file name and thread id
static void getLocalFileName (char *lfn, char *fn, int thid)
{
  sprintf (lfn, "%s.%d", fn, thid);
}

static int saveGeneralFile (int parsaving)
{
  FILE *genFile;
  int i, ss, cs;  // statistics/changes saved
  char sfn[FILENAME_MAX], cfn[FILENAME_MAX];  // statistics/changes files name
  char localfilename[FILENAME_MAX + MAXEXT + MAXLGPROC];
  time_t nowtime;
  genFile = fopen (filenameSave, "w");
  if (genFile == NULL)
    return FILEERR;
  time (&nowtime);
  ss = getStatFileName(sfn) ? 1 : 0;  // set sfn too
  cs = getChangesFileName(cfn) ? 1 : 0;  // set cfn too
  fprintf (genFile, "#General file\n#Saving time: %.24s\n\n", ctime (&nowtime));
  // cycle+1, because this is the cycle to start with when loading
  fprintf (genFile, GENFILE_STRING,
	   getTSD()->cycle + 1,
	   _nbThreads, _gSeed, _dimy, _dimx, _envType, _nbDirs,
	   getResourcePeriod(), getVisionRadius(), getVisionType(),
	   sfn, ss, cfn, cs);
  if (parsaving)
    for (i=0 ; i<_nbThreads ; i++){
      // NB: this decomposition must be the same as in initSystemPar()
      int ytop = _dimy * i / _nbThreads;
      int xtop = 0;
      getLocalFileName (localfilename, filenameSave, i);
      fprintf (genFile, "file %d %d %s\n", ytop, xtop, localfilename);  // relative to this thread
    }
  else{
    getLocalFileName (localfilename, filenameSave, 0);
    fprintf (genFile, "file 0 0 %s\n", localfilename);  // relative to (0,0)
  }
  fclose (genFile);
  return OK;
}

static void saveObstaclesPar (FILE *localfile)
{
  int y, x, ytop, ybot, xtop, xbot;
  tsd_t *ptsd = getTSD();
  I (localfile != NULL);
  fprintf (localfile, "#obstacles\n#\ty\tx\n");
  if (bParallelSavingDelayed){  // parallel saving => save only its domain
    ytop = ptsd->ytop;
    xtop = ptsd->xtop;
    ybot = ptsd->ybot;
    xbot = ptsd->xbot;
  }else{  // sequential saving (done by thread 0)
    ytop = 0;  xtop = 0;  ybot = _dimy;  xbot = _dimx;
  }
  // don't save the edges, because conflict with pmCreateEnv at pmLoadSystem
  // TODO: the avoiding of edges is not quite good
  if (ytop == 0)
    ytop++;
  if (xtop == 0)
    xtop++;
  if (ybot == _dimy)
    ybot--;
  if (xbot == _dimx)
    xbot--;
  for (y=ytop ; y<ybot ; y++)
    for (x=xtop ; x<xbot ; x++)
      if (pmgaGetSquareType (y, x) == WALLSQ)
	fprintf (localfile, OBST_ABBR "\t" OBST_STRING "\n",
		 y - ptsd->ytop, x - ptsd->xtop);
  // I assume that the compiler is intelligent and
  //   uses the same address for all the strings in the line above
  fprintf (localfile, "\n");
}

// get the id (order number in agfuncs) of the function of ag
static int getAgentfuncID (int ag)
{
  pmpfAgent myfunc;
  int i;
  I (isAlive(ag));
  myfunc = getAgentFunc(ag);
  for (i=0 ; agfuncsSave[i] ; i++)
    if (agfuncsSave[i] == myfunc)
      return i;
  return FUNCERR;
}

static int saveAgentsPar (FILE *localfile)
{
  int i;
  tsd_t *ptsd = getTSD();
  I (localfile != NULL);
  fprintf (localfile, "#agents\n#\ty\tx\tslownes\tpriorit\tload\tmaxload\tfuncInd\n");
  for (i=0 ; i<getAgentNumbermax() ; i++)
    if (isAlive(i))
      if (! bParallelSavingDelayed
	  || isInMyDomain (pmgiGetAgenty(i), pmgiGetAgentx(i))){
	int funcid = getAgentfuncID (i);
	if (funcid == FUNCERR)
	  return FUNCERR;
	fprintf (localfile, AGENT_ABBR "\t" AGENT_STRING "\n",
		 pmgiGetAgenty(i) - ptsd->ytop,
		 pmgiGetAgentx(i) - ptsd->xtop,
		 getAgentSlowness(i), getAgentPriority(i), pmgiGetAgentLoad(i),
		 pmgiGetAgentMaxload(i), funcid);
      }
  fprintf (localfile, "\n");
  return OK;
}

// get the id (order number in l2pfuncs) of the function of gradient source
static int getResourcefuncID (int type, int no)
{
  pmpfSource myfunc;
  int i;
  myfunc = getResourceFunc(type, no);
  for (i=0 ; l2pfuncsSave[i] ; i++)
    if (l2pfuncsSave[i] == myfunc)
      return i;
  return FUNCERR;
}

static int saveGradientSourcesPar (FILE *localfile)
{
  int i, type;
  tsd_t *ptsd = getTSD();
  I (localfile != NULL);
  fprintf (localfile, "#gradient sources\n#\ty\tx\ttype\tload\tinitLoa\tpot\tinitPot\tfuncInd\n");
  for (type=0 ; type<getResourceTypesNumber() ; type++)
    for (i=0 ; i<getResourceNumber(type) ; i++)
      if (! bParallelSavingDelayed
	  || isInMyDomain (getResourcey(type, i), getResourcex(type, i))){
	int funcid = getResourcefuncID (type, i);
	if (funcid == FUNCERR)
	  return FUNCERR;
	fprintf (localfile, RES_ABBR "\t" RES_STRING "\n",
		 getResourcey(type, i) - ptsd->ytop,
		 getResourcex(type, i) - ptsd->xtop,
		 type2user(type),
		 getResourceLoad(type, i), getResourceInitLoad(type, i),
		 getResourcePot(type, i), getResourceInitPot(type, i), funcid);
      }
  fprintf (localfile, "\n");
  return OK;
}

// save local file for thread number thid
int saveLocalFilePar (int thid)
{
  FILE *localFile;
  int ret;
  char localfilename[FILENAME_MAX+MAXEXT+MAXLGPROC];
  getLocalFileName (localfilename, filenameSave, thid);
  localFile = fopen (localfilename, "w");
  if (localFile == NULL)
    return FILEERR;
  fprintf (localFile, "#Local file (relative coordinates)\n");
  saveObstaclesPar (localFile);
  if ((ret = saveAgentsPar (localFile)) != OK)
    return ret;
  if ((ret = saveGradientSourcesPar (localFile)) != OK)
    return ret;
  fclose (localFile);
  return OK;
}

int pmSaveSystem (char *filename, pmpfAgent agfuncs[], pmpfSource l2pfuncs[])
{
  int i, ret, agnbfunc, gsnbfunc;
  I (strlen(filename) < FILENAME_MAX - MAXEXT);
  // get number of funcs
  for (agnbfunc=0 ; agfuncs[agnbfunc] ; agnbfunc++);
  for (gsnbfunc=0 ; l2pfuncs[gsnbfunc] ; gsnbfunc++);
  printf ("%d agfuncs & %d l2pfuncs\n", agnbfunc, gsnbfunc);
  // save parameters (NB: used in sequential saving too)
  if ((agnbfunc >= MAXFUNCS) || (gsnbfunc >= MAXFUNCS))
    return ERR;  // can store max MAXFUNCS pointers
  for (i=0 ; i<agnbfunc ; i++)
    agfuncsSave[i] = agfuncs[i];
  for (i=0 ; i<gsnbfunc ; i++)
    l2pfuncsSave[i] = l2pfuncs[i];
  strcpy (filenameSave, filename);
  // save general file
  ret = saveGeneralFile (bParallelSaving && bParallelSavingAllowed);
  if (ret != OK)
    return ret;
  // deal with local file(s)
  if (bParallelSaving && bParallelSavingAllowed){
    // parallel saving, delayed
    bParallelSavingDelayed = 1;
    return OK;
  }else
    // save the local file
    // NB: bParallelSavingDelayed is set to 0 in system.c (because pmSaveSystem can be not called)
    return saveLocalFilePar (0);
}

// --------------------------------------------------------------------

// pass FILENAME through C preprocessor, which writes to TMPFILENAME
static void preprocess (char *filename, char *tmpfilename)
{
  // gcc: -x c: C files (ignored by mipspro)
  // gcc: > or -o
  // mipspro: -P: xx.i without line directives
  //          -E: stdout with line directives
  static char format[] = "cc -x c -E %s > %s";
  char cmd[L_tmpnam + FILENAME_MAX + sizeof(format) + sizeof('\0')];
  tmpnam (tmpfilename);
  sprintf (cmd, format, filename, tmpfilename);
  VL (("Executing `%s'\n", cmd));
  system (cmd);
}

// read a local file (obstacles, agents and gradsrc)
static int loadLocalFile (localfile_t locfile,
			  pmpfAgent agfuncs[], pmpfSource l2pfuncs[])
{
  FILE *localFile;
  char tmpfilename[L_tmpnam];
  int lineno, ret, rez;
  int agnbfunc, gsnbfunc;
  int y, x, slowness, priority, agload, agmaxload, usertype;
  int load, initload, pot, initpot, idfunc;
  char field[1024], linestr[1024];  // max length of the input line
  preprocess (locfile.name, tmpfilename);
  localFile = fopen(tmpfilename, "r");
  if (localFile == NULL){
    fprintf (stderr, "Error opening/preprocessing file %s\n", locfile.name);
    return FILEERR;
  }
  lineno = 0;
  // get number of funcs
  for (agnbfunc=0 ; agfuncs[agnbfunc] ; agnbfunc++);
  for (gsnbfunc=0 ; l2pfuncs[gsnbfunc] ; gsnbfunc++);
  // scan the local file
  while (fgets (linestr, sizeof (linestr), localFile) != NULL){
    lineno++;
    rez = sscanf (linestr, "%s", field);
    if ((rez <= 0) || (field[0] == '#'))  // comments
      continue;
    if (!strcmp(field, OBST_ABBR)){       // obstacle
      if (sscanf (linestr, OBST_ABBR " " OBST_STRING, &y, &x) != 2){
	fprintf (stderr, "Error param pmPutObstacle in local file %s, line %d\n",
		 locfile.name, lineno);
	return FILEERR;
      }
      y += locfile.ytop;
      x += locfile.xtop;
      ret = pmgaGetSquareType (y, x);
      if (ret != FREESQ){
	fprintf (stderr, "Error pmPutObstacle in local file %s, line %d\n",
		 locfile.name, lineno);
	return ret;
      }
      pmPutObstacle (y, x);
    }else if (!strcmp(field, AGENT_ABBR)){  // agent
      if (sscanf (linestr, AGENT_ABBR " " AGENT_STRING,
		  &y, &x, &slowness, &priority, &agload, &agmaxload,
		  &idfunc) != 7){
	fprintf (stderr, "Error param local file %s, line %d.  Exiting\n",
		 locfile.name, lineno);
	return FILEERR;
      }
      if (idfunc >= agnbfunc){  // check idfunc
	fprintf (stderr, "Error func local file %s, line %d.  Exiting\n",
		 locfile.name, lineno);
	return FUNCERR;
      }
      // TODO: pmPutAgent uses I; to check this condition here first
      // TODO: replace 0 with the appropriate value
      putAgentAll (y+locfile.ytop, x+locfile.xtop, slowness, priority,
		   agload, agmaxload, NULL, 0, agfuncs[idfunc]);
    }else if (!strcmp(field, RES_ABBR)){  // gradient source
      if (sscanf (linestr, RES_ABBR " " RES_STRING,
		  &y, &x, &usertype, &load, &initload,
		  &pot, &initpot, &idfunc) != 8){
	fprintf (stderr, "Error param local file %s, line %d.  Exiting\n",
		 locfile.name, lineno);
	return FILEERR;
      }
      if (idfunc >= gsnbfunc){  // check idfunc
	fprintf (stderr, "Error local file %s, line %d.  Exiting\n",
		 locfile.name, lineno);
	return FUNCERR;
      }
      // TODO: pmPutAgent uses I; to check this condition here first
      putUsertypeResource (y+locfile.ytop, x+locfile.xtop, usertype,
			   load, initload, pot, initpot, l2pfuncs[idfunc]);
    }
    else{
      fprintf (stderr, "Error local file %s, line %d.  Exiting\n",
	       locfile.name, lineno);
      return FILEERR;
    }
  }
  fclose (localFile);
  remove (tmpfilename);
  printf ("local file %s succesfully loaded\n", locfile.name);
  return OK;
}

static int loadGeneralFile (char *filename)
{
  FILE *genFile;
  char tmpfilename[L_tmpnam];
  int lineno, rez, ret;
  char field[1024], linestr[1024];  // max length of the input line
  int readcycle, readnbThreads, readdimy, readdimx, readenvtype,
    readgradsrcPeriod, readvisionradius, readvisiontype, readnbdirs,
    readstatsaved, readchangessaved;
  unsigned int readgseed;
  int getcycle=0, getnbThreads=0, getgseed=0, getdimy=0, getdimx=0,
    getenvtype=0, getgradsrcPeriod=0, getvisionradius=0, getvisiontype=0,
    getnbdirs=0, getstatfilename=0, getchangesfilename=0,
    getstatsaved=0, getchangessaved=0;
  char readstatfilename[FILENAME_MAX] = "", readchangesfilename[FILENAME_MAX] = "";
  preprocess (filename, tmpfilename);
  genFile = fopen(tmpfilename, "r");
  if (genFile == NULL){
    fprintf (stderr, "Error opening/preprocessing general file %s\n", filename);
    return FILEERR;
  }
  // read general file
  lineno = 0;
  while (fgets (linestr, sizeof (linestr), genFile) != NULL){
    lineno++;
    rez = sscanf (linestr, "%s", field);  // rez==0 if blank line
    if ((rez <= 0) || (field[0] == '#'))  // skip comments
      continue;
    if (!getcycle && !strcmp(field, "cycle")){
      getcycle = 1;
      sscanf (linestr, "cycle %d", &readcycle);
    }else if (!getnbThreads && !strcmp(field, "nbThreads")){
      getnbThreads = 1;
      sscanf (linestr, "nbThreads %d", &readnbThreads);
    }else if (!getgseed && !strcmp(field, "gseed")){
      getgseed = 1;
      sscanf (linestr, "gseed %u", &readgseed);
    }else if (!getnbdirs && !strcmp(field, "nbdirs")){
      getnbdirs = 1;
      sscanf (linestr, "nbdirs %d", &readnbdirs);
    }else if (!getdimy && !strcmp(field, "dimy")){
      getdimy = 1;
      sscanf (linestr, "dimy %d", &readdimy);
    }else if (!getdimx && !strcmp(field, "dimx")){
      getdimx = 1;
      sscanf (linestr, "dimx %d", &readdimx);
    }else if (!getenvtype && !strcmp(field, "envtype")){
      getenvtype = 1;
      sscanf (linestr, "envtype %d", &readenvtype);
    }else if (!getgradsrcPeriod && !strcmp(field, "gradsrcUpdate")){
      getgradsrcPeriod = 1;
      sscanf (linestr, "gradsrcUpdate %d", &readgradsrcPeriod);
    }else if (!getvisionradius && !strcmp(field, "visionradius")){
      getvisionradius = 1;
      sscanf (linestr, "visionradius %d", &readvisionradius);
    }else if (!getvisiontype && !strcmp(field, "visiontype")){
      getvisiontype = 1;
      sscanf (linestr, "visiontype %d", &readvisiontype);
    }else if (!getstatfilename && !strcmp(field, "statfilename")){
      getstatfilename = 1;
      sscanf (linestr, "statfilename %s", readstatfilename);
    }else if (!getstatsaved && !strcmp(field, "statsaved")){
      getstatsaved = 1;
      sscanf (linestr, "statsaved %d", &readstatsaved);
    }else if (!getchangesfilename && !strcmp(field, "changesfilename")){
      getchangesfilename = 1;
      sscanf (linestr, "changesfilename %s", readchangesfilename);
    }else if (!getchangessaved && !strcmp(field, "changessaved")){
      getchangessaved = 1;
      sscanf (linestr, "changessaved %d", &readchangessaved);
    }else if (!strcmp(field, "file")){
      localfile_t lf;
      sscanf (linestr, "file %d %d %s", &lf.ytop, &lf.xtop, lf.name);
      pushstruct (&localfiles, &lf, sizeof(lf));
    }else{
      fprintf (stderr, "Error general file, line %d.  Exiting\n", lineno);
      return FILEERR;
    }
  }
  fclose (genFile);
  remove (tmpfilename);
  // only the params to pmCreateEnv are required (dimy, dimx and envtype)
  // even the local files are optional
  if (!getdimy || !getdimx || !getenvtype){
    fprintf (stderr, "Error loading general file: parameters missing\n");
    return FILEERR;
  }
  // initialize the system
  pmCreateEnv (readdimy, readdimx, readenvtype);
  if (getgseed)
    setSeed (readgseed);
  if (getnbThreads)
    pmSetThreadsNo (readnbThreads);
  if (getcycle)
    setInitialCycle (readcycle);
  if (getnbdirs)
    pmSetDirsNo (readnbdirs);
  if (getgradsrcPeriod)
    pmSetResRefreshPeriod (readgradsrcPeriod);
  if (getvisionradius)
    pmSetVisionRadius (readvisionradius);
  if (getvisiontype)
    pmSetVisionType (readvisiontype);
  if (getstatsaved && (readstatsaved == 1)){
    I (getstatfilename);  // can't have stat saved without stat file name
    ret = initStatAppend (readstatfilename);
    if (ret != OK)
      return ret;
  }else if (getstatfilename){
    ret = pmInitStat (readstatfilename);
    if (ret != OK)
      return ret;
  }
  if (getchangessaved && (readchangessaved == 1)){
    I (getchangesfilename);  // can't have changes saved without file name
    ret = initChangesAppend (readchangesfilename);
    if (ret != OK)
      return ret;
  }else if (getchangesfilename){
    ret = pmInitChanges (readchangesfilename);
    if (ret != OK)
      return ret;
  }
  printf ("general file succesfully loaded\n");
  return OK;
}

int pmLoadSystem (char *filename, pmpfAgent agfuncs[], pmpfSource l2pfuncs[])
{
  int ret, agnbfunc, gsnbfunc;
  I (strlen(filename) < FILENAME_MAX - MAXEXT);
  // get number of funcs
  for (agnbfunc=0 ; agfuncs[agnbfunc] ; agnbfunc++);
  for (gsnbfunc=0 ; l2pfuncs[gsnbfunc] ; gsnbfunc++);
  printf ("%d agfuncs & %d l2pfuncs\n", agnbfunc, gsnbfunc);
  // process general file
  ret = loadGeneralFile (filename);
  if (ret != OK)
    return ret;
  // process local files
  // TODO: the next lines have to be parallelized
  while (depthStack(&localfiles) > 0){
    localfile_t lf;
    popstruct (&localfiles, &lf, sizeof(localfile_t));
    ret = loadLocalFile (lf, agfuncs, l2pfuncs);
    if (ret != OK)
      return ret;
  }
  return OK;
}
