// This file deals with the virtual machine (mainloopPar) and the things
//   related to the system

#define INCMIGR 100      // chunk to increment migr ag structure by
#define INCGLOBALCA 100  // chunk to increment global created ag structure by

#include <stdlib.h>
#include <stdio.h>       // for stdout in L() - to be changed !!!
#include "agents.h"
#include "intern.h"
#include "lowinterface.h"
#include "nana.h"

int _nbThreads;          // *total* number of threads in the application
unsigned int _gSeed;     // the global seed
stack_t _createdAgents;

static pmpfCallback userfunc;
static stack_t migrAgents;
static int initialCycle;

void pmSetThreadsNo (int nbth)
{
  I (nbth >= 1);  // or >1 ?
  _nbThreads = nbth;
}

int pmgGetCycle (void)
{
  return getTSD()->cycle;
}

void setInitialCycle (int cycle)
{
  initialCycle = cycle;
}

void setSeed (unsigned int seed)
{
  _gSeed = seed;
}

// ----------------------------------------------------------------

// called once per simulation (possibly more than once per execution)
void pmInitSystem (unsigned int seed)
{
  _gSeed = seed;
  _nbThreads = 1;             // 1 thread by default
  _nbDirs = 4;                // 4 directions by default
  initialCycle = 0;           // by default, start with the beginning...
  initAgent ();
#ifdef WAVEPROP_BENCH
  BENCH (initResource(), -1);  // TODO: BENCH: -1: one thread only
#else
  initResource ();
#endif
  initStack (&migrAgents, INCMIGR);
  initStack (&_createdAgents, INCGLOBALCA);
  clearRandomAgent ();
  X (clearAgents());
  initSaveSys ();
}

// parallel initialization
// all TSD is initialized here, except frontChanged
static void initSystemPar (int thid)
{
  tsd_t *ptsd = getTSD();
  ptsd->id = thid;
  // for the moment, the decomposition is rectangular
  // NB: this decomposition is ideally balanced: [N*p/P, N*(p+1)/P) (add 0.5??)
  // NB: ytop and xtop of this decomposition are used also in saveGeneralFile
  ptsd->ytop = _dimy * thid / _nbThreads;
  ptsd->xtop = 0;
  ptsd->ybot = _dimy * (thid + 1) / _nbThreads;
  ptsd->xbot = _dimx;
  ptsd->cycle = initialCycle;
  initResourcePar();  // NB: this uses ptsd->id
  initConflictTablePar();
  initAgentPar();
  VL (("Decomposition for thread %d: ytop=%d, xtop=%d, ybot=%d, xbot=%d\n",
       thid, ptsd->ytop, ptsd->xtop, ptsd->ybot, ptsd->xbot));
/*   syncThreads ();  // optional: all the threads start in the same time */
}

// parallel destruction
static void endSystemPar (void)
{
  VL (("Entered endSystemPar, thread %d\n", getTSD()->id));
  freeConflictTablePar();
  freeVisionMatrixPar();
  freeResourcePar();
}

// the main loop, executed by every thread
static void mainloopPar (int thid)
{
  int ag, cf, i, y, x;
  tsd_t *ptsd = getTSD();  // safe, because it's just a pointer
  int ytop, ybot, xtop, xbot;
  VL (("Entered mainloopPar, thread %d\n", thid));
  initSystemPar (thid);  // the first thing to do, before using ptsd
  ytop = ptsd->ytop;  // NB: after initSystemPar
  ybot = ptsd->ybot;
  xtop = ptsd->xtop;
  xbot = ptsd->xbot;
#ifdef VISION_BENCH
  BENCH (allocVisionMatrixPar(), ptsd->id);
  BENCH (fillVisionMatrixPar(), ptsd->id);
  endSystemPar();
  return;  // we execute the program only for bench
#else
  allocVisionMatrixPar ();
  fillVisionMatrixPar ();
#endif
#ifdef WAVEPROP_BENCH
  BENCH (updateGradients(USEPOT), ptsd->id);  // NB: use syncThreads between if executed several times (threads exit barriers in undefined order)!!
#else
  updateGradients (USEPOT);
#endif
  while (1){
    static int userfuncret;  // shared variable
    // 0. execute user function
    if (ptsd->id == 0){  // only the thread 0 executes the userfunc
      clearStack (&_createdAgents);
      setParallelSavingDelayed (0);
      setParallelSavingAllowed (1);  // parallel saving allowed only in userfunc
      ptsd->seed = UNIQUE (ptsd->cycle, 0, 0);
      userfuncret = userfunc();
      setParallelSavingAllowed (0);
      // after userfunc, to allow him to use these info
      clearStack (&migrAgents);
      clearRandomAgent ();
      rand_r (&_gSeed);  // to assure randomness (& reproducibility)
    }
    syncThreads ();
    //   NB: here the threads join when ending the loop
    // parallel saving, if needed
    if (getParallelSavingDelayed() == 1)  // save local files
      saveLocalFilePar (ptsd->id);
    if (userfuncret == STOP)
      break;  // end of simulation
    // 1. get newpos of the agents
    // 2. fill newpos table
    // 3. fill conflicts table
    clearNewPos ();
    clearConflictTablePar ();
#ifdef DOMAIN_SCANNING
    for (y=ytop ; y<ybot ; y++)
      for (x=xtop ; x<xbot ; x++){
	if (pmgaGetSquareState(y, x) == OCCUP){  // there is an agent
	  ag = pmgaGetAgentId (y, x);
#else
    for (ag=0 ; ag<getAgentNumbermax() ; ag++)
      if (isAlive(ag)){
	y = pmgiGetAgenty (ag);
	x = pmgiGetAgentx (ag);
	if (isInMyDomain(y, x)){
#endif
	  X (testAndsetAgentplan(ag));
	  if ((ptsd->cycle % getAgentSlowness (ag)) == 0){  // it executes this cycle
	    pmAction_t act;
	    // initialise id, y, x and seed of the current agent (for local functions called in funcag)
	    ptsd->seed = UNIQUE (ptsd->cycle, y, x);
	    ptsd->crtAgentId = ag;
	    ptsd->crtAgenty = y;
	    ptsd->crtAgentx = x;
	    // the line below may fill ca structure & mark agents as killed
	    act = getAgentFunc(ag)(getAgentMemory(ag));  // execute its movement function
	    setAgentAction (ag, act);
	    switch (act.type){
	    case MOVERANDOM:
	      if (!getAgentRandomMove (ag, &act.param)){  // no free square
		setAgentActiontype (ag, MOVERANDOM_STAY);
		break;
	      }
	      setAgentActiontype (ag, MOVERANDOM_MOVE);
	      // fall through
	    case MOVE:
	      // fill ag newpos
	      planAgentMove (ag, act.param);
	      if (isInMyDomain(getAgentnewy(ag), getAgentnewx (ag)))  // this agent doesn't change its domain
		// fill newpos table and conflicts table
		addNewPos (getAgentnewy(ag), getAgentnewx (ag), ag);
	      else{  // this agent changes its domain
		X (testAndsetMAgent(ag));
		lockMAgent ();  // can I use local tables to avoid the locks?
		pushint (&migrAgents, ag);  // put this ag into the MAG structure
		unlockMAgent ();
	      }
	      break;
	    case TAKE:
	    case DROP:
	    case STAY:  // nothing to do
	      break;
	    default:
	      I (0);  // only STAY, MOVE, TAKE, DROP and MOVERANDOM are used at the moment
	    }
	  }else
	    setAgentActiontype (ag, PASS);  // this agent will not move this cycle
	}
      }
    syncThreads ();
    // deal with created agents in global stack
    // VL (("Depth of _createdAgents (agents created in other's domain): %d\n", depthStack(&_createdAgents)));
    for (i=0 ; i<depthStack(&_createdAgents) ; i++){
      agentinfo_t *agi = peekstructp (&_createdAgents, i, sizeof(agentinfo_t));
      if (isInMyDomain(agi->y, agi->x)){  // found a created agent in local domain
	// put it in the local structure
	int agindex = pushstruct (&ptsd->createdAgents, agi, sizeof(agentinfo_t));
	// fill newpos table and conflicts table
	addNewPosCA (agi->y, agi->x, agindex);
	X (agi->created = 1);  // NB: *after* pushstruct (which sets created), and with X, because only the local table needs always the field created
      }
    }
    // deal with the migrating agents
    for (i=0 ; i<depthStack(&migrAgents) ; i++){
      int newy, newx;
      ag = peekint (&migrAgents, i); // ag <- the id of the migrating agent
      newy = getAgentnewy (ag);  // only MOVE agents, (the new position is set)
      newx = getAgentnewx (ag);
      if (isInMyDomain(newy, newx)){  // found an agent entering its domain
	X (testAndclearMAgent(ag));
	// fill newpos table and conflicts table
	addNewPos (newy, newx, ag);
      }
    }
    // 4. solve all the conflicts & process created agents in conflict
    // the move agents have always less prio than created agents
    for (cf=0 ; cf<getConflictsNumberPar() ; cf++){
      int winner;
      unsigned int seed;
      int nb = getConflictItemsNumberPar(cf);  // number of move agents in conflict
      int nbCA = getConflictItemsNumberCAPar(cf);  // number of created agents in conflict
      I (nb <= _nbDirs);
      I (nbCA <= CFSIZE - _nbDirs);
      if (nbCA > 0){  // created agents exist, they have priority
	agentinfo_t *agi;
	// force move agents to stay
	for (i=0 ; i<nb ; i++){
	  ag = getConflictDataPar (cf, i);
	  if (pmgiGetAgentActiontype(ag) == MOVE)
	    setAgentActiontype (ag, STAY_CA_CONFLICT);
	  else{
	    I (pmgiGetAgentActiontype(ag) == MOVERANDOM_MOVE);
	    setAgentActiontype (ag, MOVERANDOM_STAY_CA_CONFLICT);
	  }
	}
	// solve created agents conflicts
	sortConflictStructureCAPar (cf);
	// find the winner
	agi = ((agentinfo_t *)peekstructp (&ptsd->createdAgents, getConflictDataCAPar(cf, 0), sizeof(agentinfo_t)));  // any structure, to find y,x
	seed = UNIQUE (ptsd->cycle, agi->y, agi->x);
	winner = (unsigned int)rand_r(&seed) % nbCA;  // bug glibc => cast
	for (i=0 ; i<nbCA ; i++)
	  if (i == winner)  // winner: add it to the agent structure
	    processWinnerCA (getConflictDataCAPar (cf, i));
	  else  // looser
	    processLooserCA (getConflictDataCAPar (cf, i));
      }else{  // no created agents, so can solve the move conflicts
	int nbhigh = 0, tabhigh[8]; // number and ids of the high prio agents (max 8)
	sortConflictStructurePar (cf);
	// find the winner
	ag = getConflictDataPar (cf, 0);  // any structure, to find newy,newx
	seed = UNIQUE (ptsd->cycle, getAgentnewy(ag), getAgentnewx(ag));
	for (i=0 ; i<nb ; i++)
	  if (getAgentPriority(getConflictDataPar(cf, i)) == HIGH_PR)
	    tabhigh[nbhigh++] = i;
	if (nbhigh > 0)  // high priority agents were found
	  winner = tabhigh[(unsigned int)rand_r(&seed) % nbhigh];  // glibc bug => cast
	else             // only normal priority agents
	  winner = (unsigned int)rand_r(&seed) % nb;  // glibc bug => cast
	for (i=0 ; i<nb ; i++){
	  ag = getConflictDataPar (cf, i);
	  if (i == winner){  // the winner
	    if (pmgiGetAgentActiontype(ag) == MOVE)
	      setAgentActiontype (ag, MOVE_CONFLICT);
	    else{
	      I (pmgiGetAgentActiontype(ag) == MOVERANDOM_MOVE);
	      setAgentActiontype (ag, MOVERANDOM_MOVE_CONFLICT);
	    }
	  }else{
	    if (pmgiGetAgentActiontype(ag) == MOVE)
	      setAgentActiontype (ag, STAY_CONFLICT);
	    else{
	      I (pmgiGetAgentActiontype(ag) == MOVERANDOM_MOVE);
	      setAgentActiontype (ag, MOVERANDOM_STAY_CONFLICT);
	    }
	  }
	}
      }
    }
    // process (local) created agents not in conflict
    for (i=0 ; i<depthStack(&ptsd->createdAgents) ; i++)
      if (!((agentinfo_t *)peekstructp(&ptsd->createdAgents, i, sizeof(agentinfo_t)))->created)  // if not already processed
	processWinnerCA (i);
    syncThreads ();  // because the conflicts solved by one thread may be used by another if on the frontier
    // 5. do action
    for (ag=0 ; ag<getAgentNumbermax() ; ag++)
      if (isAlive(ag)){
	int at = pmgiGetAgentActiontype (ag);
	// we can't use (y,x), because they are changed in the lines below,
	// so we use (newy,newx) for MOVE agents
	int isMove = ((at == MOVE) || (at == MOVE_CONFLICT)
		  || (at == MOVERANDOM_MOVE) || (at == MOVERANDOM_MOVE_CONFLICT));
	if ((isMove && isInMyDomain(getAgentnewy(ag), getAgentnewx(ag)))
	    || (!isMove && isInMyDomain(pmgiGetAgenty(ag), pmgiGetAgentx(ag)))){  // in my domain
	  // here, every agent is processed only once
	  if (isKilled(ag))
	    markDeadAgent (ag);
	  else{
	    I (! isKilled(ag));
	    X (testAndsetAgentmove(ag));
	    switch (pmgiGetAgentActiontype(ag)){
	    case MOVE:
	    case MOVE_CONFLICT:
	    case MOVERANDOM_MOVE:
	    case MOVERANDOM_MOVE_CONFLICT:
	      doAgentMove (ag);
	      break;
	    case TAKE:
	      doAgentTake (ag);
	      break;
	    case DROP:
	      doAgentDrop (ag);
	      break;
	    default:  // *STAY*, PASS, CREATED: nothing to do
	      break;
	    }
	  }
	}
    }
    // X (printConflictsInfo ());
    syncThreads ();  // TODO: to think if this barrier may be removed
    X (testAndclearCAgents(&ptsd->createdAgents));
    clearStack (&ptsd->createdAgents);
    X (if (ptsd->id == 0) testAndclearAgents());
    X (if (ptsd->id == 0) testAndclearCAgents(&_createdAgents));
    ptsd->cycle++;
    // update environment
#ifdef WAVEPROP_BENCH
    BENCH (updateGradients(USELOAD), ptsd->id);  // NB: use syncThreads between if executed several times (threads exit barriers in undefined order)!!
    break;  // allow execution of endSystemPar too
#endif
    if (needupdate())
      updateGradients (USELOAD);  // updateGradients ends with syncThreads
  }
  endSystemPar();
}

// this function starts the parallel execution of mainloopPar()
// rf is the user function executed at the end of every cycle
void pmRunSystem (pmpfCallback rf)
{
  int type;
  // some useful information
  VLP ((stdout, "World dim: %dx%d\n", _dimy, _dimx));
  VLP ((stdout, "Number of agents: %d\n", pmgGetAgentNumber()));
  VLP ((stdout, "Total initial load of sources: %d\n", getInitResLoad()));
  for (type=0 ; type<getResourceTypesNumber() ; type++)
    VLP ((stdout, "Potential usertype %d: sources: %d, total load: %d\n",
	  type2user(type), getResourceNumber(type), getResTotalLoad(type)));
  userfunc = rf;
#ifdef VISION_BENCH
  BENCH (allocVisionMatrix(), -1);  // TODO: BENCH: -1 is only one thread
#else
  allocVisionMatrix ();  // sequential malloc
#endif
  initResource2 ();
  createThreads (_nbThreads, mainloopPar);
}

void pmEndSystem (void)
{
  endStack (&migrAgents);
  endStack (&_createdAgents);
  freeVisionMatrix ();
  closeChangesFiles ();
  endStat ();
  freeResource ();
  freeAgent ();
  freeEnv ();
}
