// File dealing with the vision part of the library
// Radius is used (I do not see for the moment why to speak about *max* radius)

#include <stdlib.h>  // for abs
#include <stdio.h>
#include <time.h>
#include "agents.h"
#include "intern.h"
#include "lowinterface.h"
#include "nana.h"

#define MIN(x,y) (((x) < (y)) ? (x) : (y))
#define MAX(x,y) (((x) > (y)) ? (x) : (y))

// vradius==1 means the agent sees its square and the 4 (8) in its close proxim
// may be initialised here because it is initialised in pmLoadSystem also
static int vradius = 0;        // the vision radius

// the vision type (STATIC or DYNAMIC)
static int vtype = STATIC;

/* Use the fact that the vision relation is symmetric, so do only a half
   of the computes (texec/2) and use half of the memory (mem/2).
   Using half of the memory degrades the time of reading data.
   Another solution is to not reduce the memory, but only the computes,
   which does not degrade the time of reading data.
   If you use an algorithm which does not preserve the property of symmetry,
   you have to comment this line.
   Computing: the bottom half is eliminated (except the middle point).
   I prefer to stock the middle point too, because it occupies few memory,
   but it does not force me to put an if condition in pmgaaIsVisibleSquare
   in order to verify if it is from one point to the same point.
   Allocation: only the upper half is allocated (up to middle point, included)
   The bottom half was chosen to be eliminated because the other would
   have changed the indexes/variables compared to not VISION_SYMMETRIC.
*/
// the vision can use symmetry (half memory but slightly slower) or not
static int isSymmetric = 1;  // !=0 if symmetric, 0 elsewhere

/* vision matrix:
 * 1. allocation
 * 2. filling
 * 3. utilization
 * 4. verification (not done)
 * 5. deallocation
 */
static char ****visionMatrix = NULL;  // if the type char is changed, lines of code (sizeof(char) for malloc, and maybe others) have to be changed too

// compute the Euclidean distance between two squares
static int getEDistance (int y1, int x1, int y2, int x2)
{
  int dy = abs(y1-y2);  // the distance on y-axis
  int dx = abs(x1-x2);  // the distance on x-axis
  if (_envType == TORUS){  // take into account the distance through the edges
    dy = MIN (dy, _dimy - dy);
    dx = MIN (dx, _dimx - dx);
  }
  if (_nbDirs == 4)  // four directions
    return dy + dx;
  else  // _nbDirs==8
    return MAX (dy, dx);
}

// macro replacing the print of a point
// the do-while loop is used to ensure that it is equivalent to a single stmt
#define POINT(y, x)						\
        do{							\
          if (pmgaGetSquareType(normy((y)), normx((x))) == WALLSQ)	\
            /* found a square => invisible*/			\
            return 0;						\
        }while (0)
//#define POINT(y, x) printf ("%d %d\n", y, x)

#ifndef VISION_UNROLLED
// use Bresenham-like algorithm to print a line from (y1,x1) to (y2,x2)
// I think that all the parameters must be in the range [0,dim*-1]
// maybe better to use fixed point lines (if dy or dx > ...)
// return 1 if visible (no obstacle on the line), else 0.
// The difference with Bresenham is that ALL the points of the line are
//   printed, not only one per x coordinate.
// Principles of the Bresenham algorithm taken from:
//   http://www.intranet.ca/~sshah/waste/art7.html
// *vision are additions for the vision
// the line using TORUS and norm in the last line are additions for
//   the torus case (see POINT also)
static int useVisionLine (int y1, int x1, int y2, int x2)
{
  int i;               // loop counter
  int ystep, xstep;    // the step on y and x axis
  int error;           // the error accumulated during the increment
  int errorprev;       // *vision the previous value of the error variable
  int y = y1, x = x1;  // the line points
  int ddy, ddx;        // compulsory variables: the double values of dy and dx
  int dy = y2 - y1;
  int dx = x2 - x1;
  ICHECKCOORDS (y1, x1);
  ICHECKCOORDS (y2, x2);
  POINT (y1, x1);  // first point
  // NB the last point can't be added here too, because of its previous point (which has to be verified)
  if (dy < 0){
    ystep = -1;
    dy = -dy;
  }else
    ystep = 1;
  if (dx < 0){
    xstep = -1;
    dx = -dx;
  }else
    xstep = 1;
  if (_envType == TORUS){  // if torus case
    if (2*dy > _dimy){  // use the "exterior" distance
      // if 2*dy == _dimy, the distance is the same for interior and exterior -- however, the vision radius is never so big
      ystep = -ystep;
      dy = _dimy - dy;
    }
    if (2*dx > _dimx){  // use the "exterior" distance
      xstep = -xstep;
      dx = _dimx - dx;
    }
  }
  ddy = 2 * dy;  // work with double values for full precision
  ddx = 2 * dx;
  if (ddx >= ddy){  // first octant (0 <= slope <= 1)
    // compulsory initialization (even for errorprev, needed when dx==dy)
    errorprev = error = dx;  // start in the middle of the square
    for (i=0 ; i < dx ; i++){  // do not use the first point
      x += xstep;
      error += ddy;
      if (error > ddx){  // increment y if AFTER the middle ( > )
	y += ystep;
	error -= ddx;
	// *vision three cases (octant==right->right-top for directions below):
	if (error + errorprev < ddx)  // bottom square also
	  POINT (y-ystep, x);
	else if (error + errorprev > ddx)  // left square also
	  POINT (y, x-xstep);
	else{  // corner: bottom and left squares also
	  POINT (y-ystep, x);
	  POINT (y, x-xstep);
	}
      }
      POINT (y, x);
      errorprev = error;  // *vision
    }
  }else{  // the same *vision appear
    errorprev = error = dy;
    for (i=0 ; i < dy ; i++){
      y += ystep;
      error += ddx;
      if (error > ddy){
	x += xstep;
	error -= ddy;
	if (error + errorprev < ddy)
	  POINT (y, x-xstep);
	else if (error + errorprev > ddy)
	  POINT (y-ystep, x);
	else{
	  POINT (y, x-xstep);
	  POINT (y-ystep, x);
	}
      }
      POINT (y, x);
      errorprev = error;
    }
  }
  I ((normy(y) == y2) && (normx(x) == x2));  // the last point (y2,x2) has to be the same with the last point of the algorithm
  return 1;
}

#else

// This is the same function as above, except that it unrolls the loops
//   for better performance
#define BRESENHAM_LOOPS(ystep, xstep)		\
  if (ddx >= ddy){				\
    errorprev = error = dx;			\
    for (i=0 ; i < dx ; i++){			\
      x += (xstep);				\
      error += ddy;				\
      if (error > ddx){				\
	y += (ystep);				\
	error -= ddx;				\
	if (error + errorprev < ddx)		\
	  POINT (y-(ystep), x);			\
	else if (error + errorprev > ddx)	\
	  POINT (y, x-(xstep));			\
	else{					\
	  POINT (y-(ystep), x);			\
	  POINT (y, x-(xstep));			\
	}					\
      }						\
      POINT (y, x);				\
      errorprev = error;			\
    }						\
  }else{					\
    errorprev = error = dy;			\
    for (i=0 ; i < dy ; i++){			\
      y += (ystep);				\
      error += ddx;				\
      if (error > ddy){				\
	x += (xstep);				\
	error -= ddy;				\
	if (error + errorprev < ddy)		\
	  POINT (y, x-(xstep));			\
	else if (error + errorprev > ddy)	\
	  POINT (y-(ystep), x);			\
	else{					\
	  POINT (y, x-(xstep));			\
	  POINT (y-(ystep), x);			\
	}					\
      }						\
      POINT (y, x);				\
      errorprev = error;			\
    }						\
  }

// return 1 if these two points see each other - use Bresenham variant
static int useVisionLine (int y1, int x1, int y2, int x2)
{
  int i;               // loop counter
  int ystep, xstep;    // the step on y and x axis
  int error;           // the error accumulated during the increment
  int errorprev;       // *vision the previous value of the error variable
  int y = y1, x = x1;  // the line points
  int ddy, ddx;        // compulsory variables: the double values of dy and dx
  int dy = y2 - y1;
  int dx = x2 - x1;
  ICHECKCOORDS (y1, x1);
  ICHECKCOORDS (y2, x2);
  POINT (y1, x1);  // first point
  // NB the last point can't be here, because of its previous point (which has to be verified)
  if (dy < 0){
    ystep = -1;
    dy = -dy;
  }else
    ystep = 1;
  if (dx < 0){
    xstep = -1;
    dx = -dx;
  }else
    xstep = 1;
  if (_envType == TORUS){  // if torus case
    if (2*dy > _dimy){  // use the "exterior" distance
      // if 2*dy == _dimy, the distance is the same for interior and exterior -- however, the vision radius is never so big
      ystep = -ystep;
      dy = _dimy - dy;
    }
    if (2*dx > _dimx){  // use the "exterior" distance
      xstep = -xstep;
      dx = _dimx - dx;
    }
  }
  ddy = 2 * dy;  // work with double values for full precision
  ddx = 2 * dx;
  if (ystep == 1){
    if (xstep == 1){
      BRESENHAM_LOOPS(1, 1);
    }else{
      BRESENHAM_LOOPS(1, -1);
    }
  }else{
    if (xstep == 1){
      BRESENHAM_LOOPS(-1, 1);
    }else{
      BRESENHAM_LOOPS(-1, -1);
    }
  }
  I ((y == y2) && (x == x2));  // the last point (y2,x2) has to be the same with the last point of the algorithm
  return 1;
}
#endif

// return 1 if (x1,y1) and (x2,y2) see each other, else 0
// two methods: look into the matrix (filled at the beginning)
//              compute dynamically (do not use any memory for the matrix)
// the both: compute and fill the matrix dynamically (when needed) - maybe there will be many squares which do not need this matrix.  Tests: 50--75% of the squares are visited
// promising solution: have N matrix (as cache), which are filled as need is and removed by LRU (least recently used) technique when another one is needed
int pmgaaIsVisibleSquare (int y1, int x1, int y2, int x2)
{
  int yrel, xrel;  // the relative position of the second point (0..2*vradius)
  I (vradius);     // only if the vision is used
  ICHECKCOORDS (y1, x1);
  ICHECKCOORDS (y2, x2);
  yrel = normy(y2 - y1 + vradius);  // to get rid of y2==0 for example
  xrel = normx(x2 - x1 + vradius);
  I ((yrel >= 0) && (yrel < 2*vradius+1) && (xrel >= 0) && (xrel < 2*vradius+1));
  if (vtype == DYNAMIC){
    if ((_nbDirs == 4) && (getEDistance(y1, x1, y2, x2) > vradius))
      return 0;  // too far, so it is invisible
    else
      return useVisionLine (y1, x1, y2, x2);
  }
  // adjust the coordinates if necessary (i.e. if it's about the bottom half)
  if (isSymmetric
      && ((yrel > vradius) || ((yrel == vradius) && (xrel > vradius)))){
    y1 = y2;
    x1 = x2;
    yrel = 2*vradius - yrel;  // -= vradius: bug fixed!
    xrel = 2*vradius - xrel;
  }
  return visionMatrix[y1][x1][yrel][xrel];
}

// return 1 if the agent ID can see the square (y,x), else 0
// abort if the distance to the square is greater than max vision
int pmgiaIsVisibleSquare (int id, int y, int x)
{
  int yagent = pmgiGetAgenty(id);
  int xagent = pmgiGetAgentx(id);
  I (vradius);  // only if the vision is used
  ICHECKCOORDS (y, x);
  I (getEDistance(y, x, yagent, xagent) <= vradius);
  return pmgaaIsVisibleSquare(y, x, yagent, xagent);
}

int pmgmaIsVisibleSquare (int y, int x)
{
  ICHECKCOORDS (y, x);
  return pmgiaIsVisibleSquare(getTSD()->crtAgentId, y, x);
}

int pmlrIsVisibleSquare (int ry, int rx)
{
  int y = pmtraCoordy(ry);
  int x = pmtraCoordx(rx);
  return pmgmaIsVisibleSquare(y, x);
}

int pmldIsVisibleSquare (int dir)
{
  ICHECKDIR (dir);
  return pmlrIsVisibleSquare(_diry[dir], _dirx[dir]);
}

// set the vision radius
void pmSetVisionRadius (int vr)
{
  I (vr >= 0);
  vradius = vr;
}

// get the vision radius
int getVisionRadius (void)
{
  return vradius;
}

// set the vision type
void pmSetVisionType (int vt)
{
  I ((vt == STATIC) || (vt == DYNAMIC));
  vtype = vt;
}

// get the vision type
int getVisionType (void)
{
  return vtype;
}

// use the symmetry of the vision (half memory but slightly slower) or not?
// !=1 means symmetric, 0 elsewhere
void pmSetVisionSymmetry (int sym)
{
  I ((sym == 0) || (sym == 1));
  isSymmetric = sym;
}

// sequential function to allocate the vision matrix - the first two dimensions
// I can allocate all the matrix here - which has the disadvantage to need
//   contiguous memory and not to use efficiently the memory localization
//   by using the local memory (in CCNUMA case)
void allocVisionMatrix (void)
{
  int y;
  char ***p;
  if ((vradius == 0) || (vtype == DYNAMIC))  // no (pre-computed) vision
    return;
  VLP ((stdout, "Vision: %d elements (approx.) allocated\n",
	_dimy * _dimx * (isSymmetric ? (2*vradius*vradius+2*vradius+1)
			 :((2*vradius+1)*(2*vradius+1)))));
  // allocate first dimension of the matrix (0.._dimy-1)
  visionMatrix = malloc (_dimy * sizeof (char***));
  if (visionMatrix == NULL)
    abcerror ("Cannot allocate %d bytes in allocVisionMatrix.  Aborting.\n", _dimy * sizeof (char***));
  // allocate second dimension of the matrix (0.._dimx-1)
  p = malloc (_dimy * _dimx * sizeof (char**));
  if (p == NULL)
    abcerror ("Cannot allocate %d bytes in allocVisionMatrix.  Aborting.\n", _dimy * _dimx * sizeof (char**));
  for (y=0 ; y<_dimy ; y++){
    visionMatrix[y] = p;
    p += _dimx;
  }
}

// allocate the last two dimensions of the vision matrix
void allocVisionMatrixPar (void)
{
  int y, x, s, elem;
  char **p1, *p2;
  tsd_t *ptsd = getTSD();
  int ytop = ptsd->ytop, ybot = ptsd->ybot, xtop = ptsd->xtop, xbot = ptsd->xbot;
  int dy = ybot - ytop;
  int dx = xbot - xtop;  // dimensions of domain
  int ry = isSymmetric ? vradius+1 : 2*vradius+1;
  int rx = 2*vradius+1;  // dimensions of visibility field of a square
  if ((vradius == 0) || (vtype == DYNAMIC))  // no (pre-computed) vision
    return;
  I (visionMatrix);
  // allocate the third dimension (0..2*vradius or 0..vradius)
  p1 = malloc (dy * dx * ry * sizeof (char*));
  if (p1 == NULL)
    abcerror ("Cannot allocate %d bytes in allocVisionMatrixPar.  Aborting.\n", dy * dx * ry * sizeof (char*));
  for (y=ytop ; y<ybot ; y++)
    for (x=xtop ; x<xbot ; x++){
      visionMatrix[y][x] = p1;
      p1 += ry;
    }
  // allocate the 4th dimension (0..2*vradius)
  elem = ry * rx - (isSymmetric ? vradius : 0);
  p2 = malloc (dy * dx * elem * sizeof (char));
  if (p2 == NULL)
    printf ("Cannot allocate %d bytes in allocVisionMatrixPar.  Aborting.\n", dy * dx * elem * sizeof (char));
  for (y=ytop ; y<ybot ; y++)
    for (x=xtop ; x<xbot ; x++){
      for (s=0 ; s<ry ; s++){
	visionMatrix[y][x][s] = p2;
	p2 += rx;
      }
      // the last line has fewer elements
      if (isSymmetric)
	p2 -= vradius;
    }
}

// fill the vision matrix
void fillVisionMatrixPar (void)
{
  int y, x, sqy, sqx;              // env line & col, square vision lin & col
  tsd_t *ptsd = getTSD();
  int ytop = ptsd->ytop, ybot = ptsd->ybot, xtop = ptsd->xtop, xbot = ptsd->xbot;
  int lastIndex = isSymmetric ? 0 : vradius;
  if ((vradius == 0) || (vtype == DYNAMIC))  // no (pre-computed) vision
    return;
  I (visionMatrix);
  // for all its squares
  for (y=ytop ; y<ybot ; y++)
    for (x=xtop ; x<xbot ; x++)
      for (sqy=-vradius ; sqy<=lastIndex ; sqy++){
	int sqyrel = sqy+vradius;
	// fill the little matrixes
	for (sqx=-vradius ; sqx<=vradius ; sqx++){
	  if (isSymmetric && (sqy == 0) && (sqx > 0))
	    // only the upper half (the middle point included)
	    break;  // arrived also at lastIndex
	  if ((_nbDirs == 4) && (abs(sqy) + abs(sqx) > vradius))  // too far
	    visionMatrix[y][x][sqyrel][sqx+vradius] = 0;  // invisible
	  else if ((_envType == EDGES)
		   && ((y+sqy < 0) || (y+sqy >= _dimy) || (x+sqx < 0) || (x+sqx >= _dimx)))  // outside the environment
	    visionMatrix[y][x][sqyrel][sqx+vradius] = 0;  // invisible
	  else
	    visionMatrix[y][x][sqyrel][sqx+vradius] =
	      useVisionLine(y, x, normy(y+sqy), normx(x+sqx));
	}
      }
}

void freeVisionMatrixPar (void)
{
  tsd_t *ptsd = getTSD();
  int ytop = ptsd->ytop, xtop = ptsd->xtop;
  if (visionMatrix == NULL)  // not allocated
    return;
  free (visionMatrix[ytop][xtop][0]);
  free (visionMatrix[ytop][xtop]);
}

void freeVisionMatrix (void)
{
  if (visionMatrix == NULL)  // not allocated
    return;
  free (visionMatrix[0]);
  free (visionMatrix);
  visionMatrix = NULL;  // verification purposes
}
